"""This module provides support for Laplace transforms.  It acts as a
quantity for SymPy's Laplace transform.  It calculates the unilateral
Laplace transform using:

F(s) = lim_{t_0\rightarrow 0} \int_{-t_0}^{\infty} f(t) e^{-s t} dt

In comparison, SymPy uses:

F(s) = \int_{0}^{\infty} f(t) e^{-s t} dt

The latter gives 0.5 for the Laplace transform of DiracDelta(t)
whereas the former version gives 1.  Note, SymPy is inconsistent in
that it gives DiracDelta(t) for the inverse Laplace transform of 1.

Another difference with this implementation is that it will transform
undefined functions such as v(t) to V(s).

These functions are for internal use by Lcapy.

Copyright 2016--2022 Michael Hayes, UCECE

"""

from .transformer import UnilateralForwardTransformer
from .ratfun import Ratfun
from .sym import sympify, simplify, AppliedUndef
from .utils import (factor_const, scale_shift, as_sum_terms, similarity_shift,
                    expand_functions)
from .extrafunctions import rect, tri, ramp, rampstep
import sympy as sym
from warnings import warn

__all__ = ('LT', 'laplace_transform')


class LaplaceTransformer(UnilateralForwardTransformer):

    name = 'Laplace transform'

    def noevaluate(self, expr, t, s):

        t0 = sympify('t0')
        return sym.Limit(sym.Integral(expr * sym.exp(-s * t),
                                      (t, t0, sym.oo)),
                         t0, 0, dir='-')

    def check(self, expr, t, s, **kwargs):

        if expr.has(s):
            self.error('Expression depends on s')

    def key(self, expr, t, s, **kwargs):
        return (expr, t, s,
                kwargs.get('zero_initial_conditions', True))
        return expr, t, s,

    def integrate(self, expr, t, s, tmin, tmax):

        F = sym.integrate(expr * sym.exp(-s * t), (t, tmin, tmax))

        if not F.has(sym.Integral):
            return F

        if isinstance(F, sym.Integral):
            self.error('Integration failed')

        if not F.is_Piecewise:
            self.error('Expecting piecewise')

        F, cond = F.args[0]
        if F.has(sym.Integral):
            self.error('Expecting integral')

        return F

    def integrate_0minus(self, expr, t, s):

        t0 = sym.symbols('t0', negative=True, real=True)

        F = self.integrate(expr, t, s, t0, sym.oo)
        return sym.limit(F, t0, 0)

    def integrate_0(self, expr, t, s):

        return self.integrate(expr, t, s, 0, sym.oo)

    def func(self, expr, t, s, inverse=False):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        scale, shift = scale_shift(expr.args[0], t)

        # Convert v(t) to V(s), etc.
        name = expr.func.__name__
        func = sym.Function(name[0].upper() + name[1:])

        result = func(s / scale) / abs(scale)

        if shift != 0:
            result = result * sym.exp(s * shift / scale)
        return result

    def integral(self, expr, t, s):

        const, expr = factor_const(expr, t)

        if len(expr.args) != 2:
            self.error('Expecting two args')

        integrand = expr.args[0]

        if not isinstance(expr, sym.Integral):
            self.error('Expecting integral')

        if len(expr.args[1]) != 3:
            self.error('Require definite integral')

        var = expr.args[1][0]
        limits = expr.args[1][1:]
        const2, expr2 = factor_const(integrand, var)

        if (expr2.is_Function and
                expr2.args[0] == t - var and limits[0] == 0 and limits[1] == sym.oo):
            return const2 * self.term(expr2.subs(t - var, t), t, s) / s

        # Handle integral(x(tau), (tau, -oo, t))
        if not limits[0].is_positive and limits[1] == t:
            if isinstance(expr2, AppliedUndef):
                if expr2.args[0] == expr.args[1][0]:
                    return const2 * self.func(expr2, expr2.args[0], s) / s

        # Look for convolution integral
        if limits[0].is_positive:
            self.error('Cannot handle lower limit %s' % limits[0])

        if limits[1] < t:
            self.error('Cannot handle upper limit %s' % limits[1])

        if ((len(expr.args) != 2) or not expr2.is_Mul or
                not expr2.args[0].is_Function or not expr2.args[1].is_Function):
            self.error('Need integral of product of two functions')

        f1 = expr2.args[0]
        f2 = expr2.args[1]
        # TODO: apply similarity theorem if have f(a * tau) etc.

        if (f1.args[0] == var and f2.args[0] == t - var):
            F1 = self.term(f1, var, s)
            F2 = self.term(f2.subs(t - var, t), t, s)
        elif (f2.args[0] == var and f1.args[0] == t - var):
            F1 = self.term(f1.subs(t - var, t), t, s)
            F2 = self.term(f2, var, s)
        else:
            self.error('Cannot recognise convolution')

        return const2 * F1 * F2

    def derivative_undef(self, expr, t, s, zero_initial_conditions=True):

        if not isinstance(expr, sym.Derivative):
            self.error('Expecting derivative')

        if (not isinstance(expr.args[0], AppliedUndef) and
                expr.args[1][0] != t):
            self.error('Expecting function of t')

        name = expr.args[0].func.__name__
        func1 = sym.Function(name[0].upper() + name[1:])
        order = expr.args[1][1]
        result = func1(s) * s ** order

        if not zero_initial_conditions:
            # Handle initial conditions.  FIXME  use 0^- for 0.
            v = sym.Function(name)(t)
            for m in range(order):
                result -= s**(order - m - 1) * \
                    sym.Derivative(v, t, m).subs(t, 0)
        return result

    def sin_cos(self, expr, t, s):

        # Handle exp(-alpha * t) * sin(omega * t + phi) * u(t - tau)
        # The exp(-alpha * t) and u(t - tau) parts are optional.

        # Sympy sometimes has problems with this...

        factors = expr.as_ordered_factors()

        if len(factors) > 3:
            raise ValueError('Not expsin, too many factors')

        alpha = 0
        beta = 0
        m = 0
        if (factors[m].is_Function and factors[m].func is sym.exp):
            exparg = factors[m].args[0]
            alpha, beta = scale_shift(exparg, t)
            m += 1

        if not (factors[m].is_Function and factors[m].func in (sym.sin, sym.cos)):
            raise ValueError('Not expsin, no sin/cos')

        sincosarg = factors[m].args[0]
        omega, phi = scale_shift(sincosarg, t)

        if factors[m].func is sym.cos:
            phi += sym.pi / 2

        m += 1

        if len(factors) == m + 1 and not (factors[m].is_Function and factors[m].func is sym.Heaviside):
            raise ValueError('Not expsin, no Heaviside')

        tau = 0
        if len(factors) == m + 1:
            eta, zeta = scale_shift(factors[m].args[0], t)
            if eta != 1:
                raise ValueError('Need to use similarity theorem')
            tau = -zeta

            if tau.is_negative:
                tau = 0
            elif tau.is_Symbol and not tau.is_positive:
                print('Assuming %s is positive' % tau)
            elif (tau.is_Mul and len(tau.args) == 2 and
                  tau.args[0].is_negative and tau.args[1].is_Symbol):
                print('Assuming %s is positive' % -tau)
                tau = 0

        if tau != 0:
            phi += omega * tau

        E = (omega * sym.cos(phi) + (s - alpha) *
             sym.sin(phi)) / (omega**2 + (s - alpha)**2)

        if tau != 0:
            E *= sym.exp(-tau * s)
            if alpha != 0:
                E *= sym.exp(alpha * tau)

        if beta != 0:
            E = sym.exp(beta) * E

        return E

    def function(self, expr, t, s):

        expr, scale, shift = similarity_shift(expr, t)

        if shift != 0:
            # This will be handled by rewriting the function.
            return None

        if expr.func is rect:
            warn('Laplace transform ignores rect(t) for t < 0')
            return (1 - sym.exp(-s / (2 * scale))) / s

        elif expr.func is tri:
            warn('Laplace transform ignores tri(t) for t < 0')
            return 1 / s - (1 - sym.exp(-s / scale)) / (scale * s**2)

        elif expr.func is ramp:
            return scale / s**2

        elif expr.func is rampstep:
            return (1 - sym.exp(-s / scale)) / s**2

        return None

    def term(self, expr, t, s, **kwargs):

        # Unilateral LT ignores expr for t < 0 so remove Piecewise.
        if expr.is_Piecewise and expr.args[0].args[1].has(t >= 0):
            expr = expr.args[0].args[0]

        if expr.has(sym.sinh, sym.cosh, sym.tanh):
            expr = expr.rewrite(sym.exp)

        const, expr = factor_const(expr, t)

        terms = expr.expand(deep=False).as_ordered_terms()
        if len(terms) > 1:
            result = 0
            for term in terms:
                result += self.term(term, t, s, **kwargs)
            return const * result

        tsym = sympify(str(t))
        expr = expr.replace(tsym, t)

        if expr.has(sym.Integral):
            return self.integral(expr, t, s) * const

        if expr.has(sym.sin, sym.cos):
            try:
                return self.sin_cos(expr, t, s) * const
            except:
                pass

        if expr.has(AppliedUndef):

            if (expr.is_Mul and len(expr.args) == 2 and
                isinstance(expr.args[0], sym.DiracDelta) and
                    isinstance(expr.args[1], (AppliedUndef, sym.Subs))):
                return expr.args[1]

            if expr.has(sym.Derivative):
                return self.derivative_undef(expr, t, s, **kwargs) * const

            factors = expr.as_ordered_factors()
            if len(factors) == 1:
                return const * self.func(factors[0], t, s)
            elif len(factors) > 2:
                self.error('Cannot handle product')

            foo = factors[1]
            if foo.is_Function and foo.func == sym.exp and foo.args[0].has(t):
                scale, shift = scale_shift(foo.args[0], t)
                if shift == 0:
                    result = self.func(factors[0], t, s)
                    return const * result.subs(s, s - scale)
            self.error('Cannot handle product')

        if expr.is_Function and expr.args[0].has(t):
            result = self.function(expr, t, s)
            if result is not None:
                return result * const
            expr = expand_functions(expr, t)

        if expr.has(sym.Heaviside(t)):
            return self.integrate_0(expr.replace(sym.Heaviside(t), 1), t, s) * const

        if expr.has(sym.DiracDelta) or expr.has(sym.Heaviside):
            try:
                return self.integrate_0minus(expr, t, s) * const
            except ValueError:
                pass

        return self.integrate_0(expr, t, s) * const


laplace_transformer = LaplaceTransformer()


def laplace_transform(expr, t, s, evaluate=True, **kwargs):
    """Compute unilateral Laplace transform of expr with lower limit 0-.
    """

    return laplace_transformer.transform(expr, t, s,
                                         evaluate=evaluate,
                                         **kwargs)


def LT(expr, t, s, evaluate=True, **kwargs):
    """Compute unilateral Laplace transform of expr with lower limit 0-.
    """

    return laplace_transformer.transform(expr, t, s,
                                         evaluate=evaluate,
                                         **kwargs)



from lcapy import Circuit

cct = Circuit("""
V 1 0 step 20
R 1 2 10
C 2 0 1e-4
""")

vc = cct.C.v

import numpy as np
t = np.linspace(0, 0.01, 1000)

from matplotlib.pyplot import subplots, savefig
fig, ax = subplots(1)
vc.plot(t, axes=ax)

savefig('circuit-VRC2-vc.png')



"""This module contains functions for simplifying expressions.

Copyright 2020--2021 Michael Hayes, UCECE

"""

from sympy import Add, Mul, DiracDelta, Heaviside, Integral, re, im
from sympy import oo, sin, cos, sqrt, atan2, pi, Symbol, solve, Min, Max
from sympy import cosh, sinh, tanh, exp
from .extrafunctions import UnitStep, UnitImpulse, rect, dtrect


def simplify_dirac_delta_product_term(expr):
    """ Simplify f(t) * DiracDelta(t) to f(0) * DiracDelta(t)."""

    if not expr.has(DiracDelta):
        return expr

    def query(expr):

        return expr.is_Mul and expr.has(DiracDelta)

    def value(expr):

        arg = None
        factors = expr.args
        dirac = None
        parts = []
        for factor in factors:
            if factor.is_Function and factor.func == DiracDelta:
                arg = factor.args[0]
                dirac = factor
            else:
                parts.append(factor)

        if arg is None or not arg.has(Symbol):
            return expr
        results = solve(arg, dict=True)
        # Note, the eval method is called for functions.
        const = Mul(*parts).subs(results[0])
        return const * dirac

    return expr.replace(query, value)


def simplify_dirac_delta_product(expr, expand=False):
    """Simplify f(t) * DiracDelta(t) to f(0) * DiracDelta(t)."""

    if not expr.has(DiracDelta):
        return expr

    # Could also convert delta(a * t) to delta(t) / a

    if not expand:
        return simplify_dirac_delta_product_term(expr)

    terms = expr.expand().as_ordered_terms()

    return Add(*[simplify_dirac_delta_product_term(term) for term in terms])


def simplify_dirac_delta(expr, var=None):

    if not expr.has(DiracDelta):
        return expr

    expr = simplify_dirac_delta_product(expr)
    if var is not None:

        # Convert delta(a * t) to delta(t) / a
        expr = expr.expand(diracdelta=True, wrt=var)
    return expr


# def simplify_heaviside_product(expr):
#
#     heaviside_products = []
#
#     def pre(expr):
#         if (expr.is_Mul and expr.args[0].func == Heaviside and
#               expr.args[1].func == Heaviside):
#             heaviside_products.append(expr)
#
#         for arg in expr.args:
#             pre(arg)
#
#     pre(expr)
#
#     for product in heaviside_products:
#         # TODO
#         pass
#
#     return expr


def simplify_power(expr):

    powers = []

    def pre(expr):
        if (expr.is_Pow and expr.args[0].func in (Heaviside, UnitStep, rect, dtrect) and
                expr.args[1].is_constant):
            powers.append(expr)

        for arg in expr.args:
            pre(arg)

    pre(expr)

    for power in powers:
        expr = expr.replace(power, power.args[0])
    return expr


def simplify_heaviside_integral(expr):

    if not expr.has(Integral):
        return expr

    def query(expr):

        if not isinstance(expr, Integral):
            return False
        return expr.has(Heaviside) or expr.has(UnitStep)

    def value(expr):

        integrand = expr.args[0]
        var = expr.args[1][0]
        lower_limit = expr.args[1][1]
        upper_limit = expr.args[1][2]

        # Rewrite integral limits if Heaviside is a factor of the
        # integrand.

        result = 1
        for factor in integrand.as_ordered_factors():
            if isinstance(factor, (Heaviside, UnitStep)):
                arg = factor.args[0]
                if arg == var:
                    lower_limit = Max(lower_limit, 0)
                    factor = 1
                elif arg == -var:
                    upper_limit = Min(upper_limit, 0)
                    factor = 1
                elif (arg.is_Add and arg.args[1].is_Mul and
                      arg.args[1].args[0] == -1 and arg.args[1].args[1] == var):
                    upper_limit = Min(upper_limit, arg.args[0])
                    # Cannot remove Heaviside function in general.

            result *= factor

        ret = Integral(result, (var, lower_limit, upper_limit))
        return ret

    expr = expr.replace(query, value)

    return expr


def simplify_heaviside_scale(expr, var):

    terms = expr.as_ordered_terms()
    if len(terms) > 1:
        result = 0
        for term in terms:
            result += simplify_heaviside_scale(term, var)
        return result

    def query(expr):

        return expr.is_Function and expr.func in (Heaviside, UnitStep)

    def value(expr):

        arg = expr.args[0]

        if not arg.as_poly(var).is_linear:
            return expr

        arg = arg.expand()
        a = arg.coeff(var, 1)
        b = arg.coeff(var, 0)
        if a == 0:
            return expr

        return expr.func(var + (b / a).cancel())

    return expr.replace(query, value)


def simplify_heaviside(expr, var=None):

    if not expr.has(Heaviside) and not expr.has(UnitStep):
        return expr

    expr = simplify_heaviside_integral(expr)
    expr = simplify_power(expr)
    if var is not None:
        expr = simplify_heaviside_scale(expr, var)
    return expr


def simplify_rect(expr, var=None):

    if not expr.has(rect) and not expr.has(dtrect):
        return expr

    expr = simplify_power(expr)
    return expr


def simplify_sin_cos(expr, as_cos=False, as_sin=False):

    if not (expr.has(sin) and expr.has(cos)):
        return expr

    terms = expr.expand().as_ordered_terms()

    rest = 0
    cos_parts = []
    sin_parts = []

    for term in terms:
        if term.has(sin):
            sin_parts.append(term)
        elif term.has(cos):
            cos_parts.append(term)
        else:
            rest += term

    if cos_parts == [] or sin_parts == []:
        return expr

    result = 0

    # Use list to make copy for iterator.
    for cos_part in list(cos_parts):
        cfactors = cos_part.expand().as_ordered_factors()

        for sin_part in list(sin_parts):
            sfactors = sin_part.expand().as_ordered_factors()

            commonfactors = []
            for factor in cfactors:
                if factor in sfactors:
                    commonfactors.append(factor)

            for factor in commonfactors:
                sfactors.remove(factor)
                cfactors.remove(factor)

            cos_factor = None
            sin_factor = None
            for cfactor in cfactors:
                if cfactor.has(cos):
                    cos_factor = cfactor
                    break

            for sfactor in sfactors:
                if sfactor.has(sin):
                    sin_factor = sfactor
                    break

            if cos_factor is None or sin_factor is None:
                continue

            if cos_factor.args[0] != sin_factor.args[0]:
                continue

            cfactors.remove(cos_factor)
            sfactors.remove(sin_factor)

            c = Mul(*cfactors)
            s = Mul(*sfactors)
            A = sqrt(c * c + s * s) * Mul(*commonfactors)
            phi = atan2(s, c)

            if as_sin:
                result += A * \
                    sin(cos_factor.args[0] - phi + pi / 2, evaluate=False)

            if as_cos:
                result += A * cos(cos_factor.args[0] - phi, evaluate=False)

            # SymPy will choose sin or cos as convenient.
            if not as_sin and not as_cos:
                result += A * cos(cos_factor.args[0] - phi)

            cos_parts.remove(cos_part)
            sin_parts.remove(sin_part)
            break

    result = Add(result, rest, *cos_parts, *sin_parts)
    return result


def simplify_unit_impulse(expr, var=None):

    if not expr.has(UnitImpulse):
        return expr

    def query(expr):

        return expr.is_Function and expr.func is UnitImpulse

    def value(expr):

        arg = expr.args[0]

        if not arg.as_poly(var).is_linear:
            return expr

        arg = arg.expand()
        a = arg.coeff(var, 1)
        b = arg.coeff(var, 0)
        if a == 0:
            return expr

        return expr.func(var + (b / a).cancel())

    return expr.replace(query, value)


def simplify_conjugates(expr):

    factors = expr.as_ordered_factors()

    if len(factors) > 1:
        sfactors = []
        for factor in factors:
            sfactors.append(simplify_conjugates(factor))
        return Mul(*sfactors)

    terms = expr.expand().as_ordered_terms()

    sterms = []
    for m, term in enumerate(terms):
        if term == 0:
            continue
        cterm = term.conjugate()

        for m1, term1 in enumerate(terms[m + 1:]):
            if cterm == term1:
                terms[m1 + m + 1] = 0
                term = 2 * re(term)
                break
            elif cterm == -term1:
                terms[m1 + m + 1] = 0
                term = 2 * im(term)
                break

        sterms.append(term)

    return Add(*sterms)


def expand_hyperbolic_trig(expr):

    # Note, rewrite(exp) does this except it also converts s**2 to
    # exp(2 * log(s)).

    def query(expr):
        return expr.is_Function and expr.func in (cosh, sinh, tanh)

    def value(expr):
        arg = expr.args[0]

        if expr.func == cosh:
            return exp(arg) + exp(-arg)
        elif expr.func == sinh:
            return exp(arg) - exp(-arg)
        elif expr.func == tanh:
            return (exp(arg) - exp(-arg)) / (exp(arg) + exp(-arg))
        else:
            raise RuntimeError('Internal error')

    return expr.replace(query, value)


# Could have simplify_sum for things like Sum(0**m * x(m), (m, 0, oo))
# that simplifies to x(0).



from lcapy import *
import sympy as sym
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_DTFT(self):

        x = delta(n)
        X = x.DTFT(images=0)
        self.assertEqual(X, 1, "delta(n)")
        self.assertEqual(X.IDTFT(), x, "delta(n)")

        y = 2 * delta(n)
        Y = y.DTFT(images=0)
        self.assertEqual(Y, 2, "2 * delta(n)")
        self.assertEqual(Y.IDTFT(), y, "2 * delta(n)")

        Xoo = x.DTFT(images=oo)
        Yoo = y.DTFT(images=oo)
        Zoo = Xoo + Yoo

        self.assertEqual(Zoo.remove_images(), X + Y, "remove_images")
        self.assertEqual(Zoo.IDTFT(), x + y, "sum IDTFT")

        x = 0 * n + 2
        self.assertEqual(x.DTFT(images=0), 2 * delta(f) / dt, "2")
        self.assertEqual(x.DTFT(images=0).IDTFT(), x, "2")

        x = nexpr('x(n)')
        self.assertEqual(x.DTFT(images=0), fexpr('X(f)'), "x(n)")
        self.assertEqual(x.DTFT(images=0).IDTFT(), x, "x(n)")

        x = nexpr('2 * x(n)')
        self.assertEqual(x.DTFT(images=0), fexpr('2 * X(f)'), "2 * x(n)")
        self.assertEqual(x.DTFT(images=0).IDTFT(), x, "2 * x(n)")

        x = nexpr('x(2 * n)')
        self.assertEqual(x.DTFT(images=0), fexpr('X(f / 2) / 2'), "x(2 * n)")
        self.assertEqual(x.DTFT(images=0).IDTFT(), x, "x(2 * n)")

        x = sign(n)
        X = x.DTFT(Omega)
        X1 = 2 * exp(j * Omega) / (exp(j * Omega) - 1)

        self.assertEqual(X, X1, "sign(n).DTFT(Omega)")

        m = symbol('m', integer=True)
        x = nexpr('x(n)')
        X = x.DTFT(f)
        # Produces infinite recursion on replacement
        #Y = Sum(fexpr('X(f)'), (m, -oo, oo)).replace(f, f - m / dt)

        m = symbol('m', integer=True)
        # Produces infinite recursion on comparison
        #self.assertEqual(X, Y, "x(n).DTFT(F)")



"""This module provides the DiscreteTimeDomainSequence class to
represent discrete-time domain sequences.

Copyright 2021 Michael Hayes, UCECE

"""

from .domains import DiscreteTimeDomain
from .sequence import Sequence
from .sym import nsym

__all__ = ('nseq', )


class DiscreteTimeDomainSequence(DiscreteTimeDomain, Sequence):
    """Discrete-time domain sequence."""

    var = nsym
    domain = 'discrete time sequence'

    def DFT(self):
        """Calculate DFT and return as sequence."""

        from sympy import exp
        from .sym import j, pi
        from .kexpr import k

        results = []
        vals = self.vals
        N = len(vals)
        for ki in range(N):
            result = 0
            for ni in range(N):
                result += vals[ni] * exp(-2 * j * pi * self.n[ni] * ki / N)
            result = result.change(result, domain='discrete fourier')
            results.append(result)

        return self.change(results, domain='discrete fourier sequence')

    def ZT(self):
        """Calculate z-transform and return as sequence."""

        from .zexpr import z

        results = []
        vals = self.vals
        N = len(vals)
        for ni in range(N):
            result = z**(-ni) * vals[ni].expr
            result = result.change(result, domain='Z')
            results.append(result)
        return self.change(results, domain='Z sequence')


def nseq(arg, ni=None, origin=None):
    """Create a discrete-time domain Sequence from a tuple, list, ndarray, or str.

    >>> a = nseq((1, 2, 3))

    The sequence indices are specified with the optional `ni` argument.
    For example:

    >>> a = nseq((1, 2, 3, 4), (-1, 0, 1, 2))

    If the `ni` argument is not specified, the sequence indices
    are enumerated from 0.

    With a string argument, an underscore indicates the zero sequence
    index:

    >>> a = nseq('{1, _2, 3, 4}')

    The sequence indices can be found using the `n` attribute.
    This returns a list.

    >>> a = nseq('{1, _2, 3, 4}').n
    [-1, 0, 1, 2]
    """

    return DiscreteTimeDomainSequence(arg, ni, origin)


from .expressionclasses import expressionclasses  # nopep8

expressionclasses.register('discrete time sequence',
                           DiscreteTimeDomainSequence)



"""This module provides base transformer classes.

Copyright 2020--2022 Michael Hayes, UCECE

"""

from sympy import Eq, Symbol, Piecewise, S, Heaviside
from .state import state
from .sym import miscsymbol
from .utils import factor_const, remove_images
from .extrafunctions import UnitStep
from warnings import warn


class Transformer(object):

    name = 'undefined'
    is_inverse = False
    is_bilateral = False

    def __init__(self):

        self.cache = {}
        self.expr = None

    def clear_cache(self):

        self.cache = {}

    def error(self, message=''):
        if message == '':
            raise ValueError('Could not compute %s for %s' %
                             (self.name, self.expr))
        raise ValueError('Could not compute %s for %s: %s' %
                         (self.name, self.expr, message))

    def debug(self, message=''):
        if self._debug:
            print(self.name.capitalize() + ': ' + message)

    def simplify_term(self, expr, var):
        return expr

    def rewrite(self, expr, var):
        return expr

    def transform(self, expr, var, conjvar, evaluate=True, **kwargs):

        # Squirrel away original expression for error messages
        self.expr = expr
        self._debug = kwargs.pop('debug', False)

        if expr.is_Equality:
            return Eq(self.transform(expr.args[0], var, conjvar, evaluate, **kwargs),
                      self.transform(expr.args[1], var, conjvar, evaluate, **kwargs))

        # Handle Lcapy symbols for the variable and conjugate variable
        try:
            var = var.expr
        except:
            pass

        try:
            conjvar = conjvar.expr
        except:
            pass

        # The variable may have been created with different attributes,
        # say when using sym.sympify('DiracDelta(t)') since this will
        # default to assuming that t is complex.  So if the symbol has the
        # same representation, convert to the desired one.
        svar = Symbol(str(var))
        expr = expr.replace(svar, var)

        self.check(expr, var, conjvar, **kwargs)

        return self.doit(expr, var, conjvar, evaluate, **kwargs)

    def dummy_var(self, expr, dummy='m', level=0, **kwargs):
        """Create a dummy variable."""

        # Avoid i, j to reduce confusion with imaginary number
        # Avoid k, n since they are domain variables
        dummies = [dummy] + ['m', 'p', 'q', 'o']

        for dummy in dummies:
            if level == 0:
                nu = miscsymbol(dummy, **kwargs)
            else:
                nu = miscsymbol(dummy + '_%d' % level, **kwargs)
            if not expr.has(nu):
                return nu
        raise self.error(
            'Dummy variable conflict with symbols: %s' % ', '.join(dummies))


class BilateralForwardTransformer(Transformer):

    is_bilateral = True
    is_inverse = False

    def doit(self, expr, var, conjvar, evaluate=True, cache=True, **kwargs):

        const, expr = factor_const(expr, var)

        key = self.key(expr, var, conjvar, **kwargs)
        if cache and key in self.cache:
            return const * self.cache[key]

        expr = self.rewrite(expr, var)

        terms = expr.as_ordered_terms()
        result = 0

        for term in terms:
            sterm = self.simplify_term(term, var)
            ret = self.term(sterm, var, conjvar, **kwargs)
            result += ret

        self.cache[key] = result
        return const * result


class BilateralInverseTransformer(BilateralForwardTransformer):

    is_inverse = True


class UnilateralForwardTransformer(Transformer):

    is_bilateral = False

    def remove_heaviside(self, expr, var):

        rest = S.One
        for factor in expr.as_ordered_factors():
            if (factor.is_Function and
                factor.func in (Heaviside, UnitStep) and
                    factor.args[0] == var):
                pass
            else:
                rest *= factor
        return rest

    def simplify_term(self, expr, var):

        return self.remove_heaviside(expr, var)

    def doit(self, expr, var, conjvar, evaluate=True, **kwargs):

        # Unilateral transforms ignore expr for t < 0 so remove Piecewise.
        if expr.is_Piecewise and expr.args[0].args[1].has(var >= 0):
            expr = expr.args[0].args[0]

        if not evaluate:
            return self.noevaluate(expr, var, conjvar)

        const, expr = factor_const(expr, var)

        key = self.key(expr, var, conjvar, **kwargs)
        if key in self.cache:
            return const * self.cache[key]

        expr = self.rewrite(expr, var)

        terms = expr.as_ordered_terms()
        result = 0

        for term in terms:
            sterm = self.simplify_term(term, var)
            ret = self.term(sterm, var, conjvar, **kwargs)
            result += ret

        self.cache[key] = result
        return const * result


class UnilateralInverseTransformer(Transformer):

    is_bilateral = False
    is_inverse = True

    def make(self, var, const, cresult, uresult, **kwargs):

        result = const * (cresult + uresult)

        # The original signal can only be recovered from an inverse
        # unilateral Laplace transform if we have the region of
        # convergence.  In general, the result is unknown for t < 0
        # unless we know the result is causal.

        if not kwargs.get('causal', False):

            if uresult != 0:
                # Cannot determine result for var < 0
                result = Piecewise((result, var >= 0))

        return result

    def doit(self, expr, var, conjvar, evaluate=True, **kwargs):

        if not evaluate:
            return self.noevaluate(expr, var, conjvar)

        const, expr = factor_const(expr, var)

        key = self.key(expr, var, conjvar, **kwargs)
        if key in self.cache:
            return self.make(conjvar, const, *self.cache[key], **kwargs)

        expr = self.rewrite(expr, var)

        terms = expr.as_ordered_terms()

        uresult = 0
        cresult = 0

        for term in terms:
            sterm = self.simplify_term(term, var)
            cterm, uterm = self.term(sterm, var, conjvar, **kwargs)
            cresult += cterm
            uresult += uterm

        self.cache[key] = cresult, uresult
        return self.make(conjvar, const, *self.cache[key], **kwargs)



from .quantity import Quantity


class TransferMixin(Quantity):

    quantity = 'transfer'
    quantity_label = 'Transfer function'
    quantity_units = ''
    is_always_causal = True
    # FIXME, non causal transfer functions can be constructed
    # but they won't result from circuit analysis.
    is_transfer = True
    is_ratio = True



"""This module defines special symbols f, s, t, and omega.

Copyright 2014--2021 Michael Hayes, UCECE

"""

# The following domain variables are Lcapy expressions.  For example, f is an
# FourierDomainExpression object wrapping the fsym symbol.
from .fexpr import f
from .texpr import t
from .sexpr import s
from .omegaexpr import omega
from .normomegaexpr import Omega
from .normfexpr import F
from .nexpr import n
from .kexpr import k
from .zexpr import z

# Import common SymPy symbols.
from .sym import pi, j, oo, inf, one, omega0sym

# Perhaps have class for domainconstants such as omega0, t0, f0, etc.
omega0 = omega0sym

jomega = j * omega
jw = jomega

jomega0 = j * omega0
jw0 = jomega0

# This represents an arbitrary angular frequency
w = omega
# This represents a specific angular frequency and is assumed to be positive
w0 = omega0

W = Omega

domain_vars = [t, f, s, omega, omega0, Omega, F, n, k, z, jw, jw0]
domain_var_ids = [id(var) for var in domain_vars]



"""This module maintains a registry of symbols.

Copyright 2022 Michael Hayes, UCECE

"""

from sympy import Symbol
from .attrdict import AttrDict
from warnings import warn

# Cannot add attributes to SymPy symbols since the Symbol
# class defines __slots__ = ().  So perhaps use facade for
# symbols that includes the symbol kind?
kinds = {}


class SymbolRegistry(AttrDict):
    """This maintains a registry of symbols so that there is only
    a single symbol of a given name.

    The symbol kinds are:
    'domain' defined by Lcapy for domain variables
    'misc' defined by Lcapy for misc. variables, such as nu
    'user' defined by the user using `symbol()` or `symbols()`
    'expr' defined by the user using `expr()`
    """

    def register(self, name, symbol, kind='user'):
        from .state import state

        if state.notify_symbol_add:
            print("Adding symbol '%s'" % name)

        self[name] = symbol
        kinds[name] = kind
        return symbol

    def add(self, name: str, kind='user', override=True,
            force=False, **assumptions):

        if name in self:
            symbol = self[name]

            if not override:
                return symbol

            if kinds[name] == 'domain' and not force:
                raise ValueError(
                    'Cannot override domain symbol %s without force=True' % name)

        if assumptions == {}:
            assumptions['positive'] = True
            # Note this implies that imag is False.   Also note that all
            # reals are considered complex (but with a zero imag part).

        elif 'positive' in assumptions:
            if not assumptions['positive']:
                assumptions.pop('positive')

        symbol = Symbol(name, **assumptions)
        return self.register(name, symbol, kind)

    def delete(self, name):

        self.pop(name)

    def add_domain(self, name: str, **assumptions):

        return self.add(name, kind='domain', **assumptions)

    def add_misc(self, name: str, **assumptions):

        return self.add(name, kind='misc', **assumptions)

    def add_user(self, name: str, **assumptions):

        return self.add(name, kind='user', **assumptions)

    def add_from_expr(self, expr):
        """Add all the symbols in an expression."""

        for symbol in expr.atoms(Symbol):
            name = str(symbol)
            self.register(name, symbol, kind='expr')

    def delete_user(self):
        """Delete user symbols."""

        for name in self.user().keys():
            self.delete(name)

    def by_kind(self, kind):

        syms = self.__class__()
        for name, sym in self.items():
            if kinds[name] == kind:
                syms[name] = sym
        return syms

    def user(self):
        """Return user defined symbols."""

        return self.by_kind('user')

    def expr(self):
        """Return expr defined symbols."""

        return self.by_kind('expr')

    def delete_expr(self):
        """Delete expr symbols."""

        for name in self.expr().keys():
            self.delete(name)

    def clear(self):

        self.delete_expr()
        self.delete_user()

    def lookup(self, name):

        new = name
        if not isinstance(name, str):
            name = str(name)

        # Replace symbol names with symbol definitions to
        # avoid problems with real or positive attributes.
        try:
            return self[name]
        except:
            from .state import state

            # This can occur when trying to substitute an expression
            # rather than a symbol, for example, when making state-space.
            if state.warn_unknown_symbol:
                warn('Lcapy does not know symbol %s' % name)
            if state.break_unknown_symbol:
                import pdb
                pdb.set_trace()

        return new

    def kind(self, name):
        """Return the symbol kind or 'unknown' if unknown."""

        try:
            return kinds[name]
        except:
            return 'unknown'



"""This module provides the DiscreteFourierDomainSequence class to
represent discrete-Fourier domain sequences.

Copyright 2021 Michael Hayes, UCECE

"""

from .domains import DiscreteFourierDomain
from .sequence import Sequence
from .sym import ksym

__all__ = ('kseq', )


class DiscreteFourierDomainSequence(DiscreteFourierDomain, Sequence):
    """Discrete-Fourier domain sequence."""

    var = ksym
    domain = 'discrete fourier sequence'

    def IDFT(self):
        """Calculate IDFT and return as sequence."""

        from sympy import exp
        from .sym import j, pi
        from .nexpr import n

        results = []
        vals = self.vals
        N = len(vals)
        for ni in range(N):
            result = 0
            for ki in range(N):
                result += vals[ki] * exp(2 * j * pi * ni * self.n[ki] / N)
            result = result.change(result, domain='discrete time')
            results.append(result / N)

        return self.change(results, domain='discrete time sequence')


def kseq(arg, ni=None, origin=None):
    """Create a discrete-Fourier domain Sequence from a tuple, list, ndarray, or str.

    >>> a = kseq((1, 2, 3))

    The sequence indices are specified with the optional `ni` argument.
    For example:

    >>> a = kseq((1, 2, 3, 4), (-1, 0, 1, 2))

    If the `ni` argument is not specified, the sequence indices
    are enumerated from 0.

    With a string argument, an underscore indicates the zero sequence
    index:

    >>> a = kseq('{1, _2, 3, 4}')

    The sequence indices can be found using the `n` attribute.
    This returns a list.

    >>> a = kseq('{1, _2, 3, 4}').n
    [-1, 0, 1, 2]
    """

    return DiscreteFourierDomainSequence(arg, ni, origin)


from .expressionclasses import expressionclasses  # nopep8

expressionclasses.register('discrete fourier sequence',
                           DiscreteFourierDomainSequence)



"""This module provides the DiscreteTimeDomainExpression class to
represent discrete-time expressions.

Copyright 2020--2022 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import DiscreteTimeDomain
from .functions import UnitStep
from .sym import oo, fsym
from .sym import nsym, ksym, zsym, dt
from .ztransform import ztransform
from .dft import DFT
from .seqexpr import SequenceExpression
from .nseq import DiscreteTimeDomainSequence, nseq
from sympy import Sum, summation, limit


__all__ = ('nexpr', )


class DiscreteTimeDomainExpression(DiscreteTimeDomain, SequenceExpression):
    """Discrete-time expression or symbol."""

    var = nsym
    seqcls = DiscreteTimeDomainSequence

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)
        if 'integer' not in assumptions:
            assumptions['real'] = True

        super(DiscreteTimeDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr
        if check and expr.has(zsym) and not expr.has(Sum):
            raise ValueError(
                'n-domain expression %s cannot depend on z' % expr)
        if check and expr.has(ksym) and not expr.has(Sum):
            raise ValueError(
                'n-domain expression %s cannot depend on k' % expr)

    def _mul_compatible_domains(self, x):

        if self.domain == x.domain:
            return True

        return x.is_constant_domain

    def _div_compatible_domains(self, x):

        if self.domain == x.domain:
            return True

        return x.is_constant_domain

    def as_expr(self):
        return DiscreteTimeDomainExpression(self)

    def differentiate(self):
        """First order difference."""

        result = (self.expr - self.subs(n - 1).expr) / dt
        return self.__class__(result, **self.assumptions)

    def integrate(self):
        """First order integration."""

        from .sym import miscsymbol
        from .utils import factor_const
        from .extrafunctions import UnitImpulse
        from .functions import u

        # TODO, get SymPy to optimize this case.
        expr = self.expr
        const, expr = factor_const(expr, nsym)
        if expr.is_Function and expr.func == UnitImpulse:
            return dt * u(expr.args[0]) * const

        msym = miscsymbol('m', integer=True)
        result = dt * summation(self.subs(msym).expr, (msym, -oo, nsym))

        return self.__class__(result, **self.assumptions)

    def ztransform(self, evaluate=True, **assumptions):
        """Determine one-sided z-transform."""

        assumptions = self.assumptions.merge_and_infer(self, **assumptions)
        result = ztransform(self.expr, self.var, zsym, evaluate)
        return self.change(result, domain='Z', **assumptions)

    def ZT(self, **assumptions):
        return self.ztransform(**assumptions)

    def plot(self, ni=None, **kwargs):
        """Plot the sequence.  If `ni` is not specified, it defaults to the
        range (-20, 20).  `ni` can be a vector of specified sequence
        indices, a tuple specifing the range, or a constant specifying
        the maximum value with the minimum value set to 0.

        kwargs include:
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label
        ylabel - the y-axis label
        xscale - the x-axis scaling, say for plotting as ms
        yscale - the y-axis scaling, say for plotting mV
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned.

        """

        if ni is None:
            ni = (-20, 20)

        from .plot import plot_sequence
        return plot_sequence(self, ni, **kwargs)

    def initial_value(self):
        """Determine value at n = 0."""

        return self.subs(0)

    def final_value(self):
        """Determine value at n = oo."""

        return self.__class__(limit(self.expr, self.var, oo))

    def DFT(self, N=None, evaluate=True, piecewise=False):
        """Determine DFT.

        `N` needs to be a positive integer symbol or a str specifying
        the extent of the DFT.  By default `N` is defined as 'N'."""

        from .sym import miscsymbol

        if N is None:
            N = miscsymbol('N', integer=True, positive=True)
        elif isinstance(N, str):
            N = miscsymbol(N, integer=True, positive=True)

        result = DFT(self.expr, nsym, ksym, N, evaluate=evaluate,
                     piecewise=piecewise)
        result = self.change(result, domain='discrete fourier')
        result = result.simplify_unit_impulse()
        return result

    def delay(self, m):
        """Delay signal by m samples."""

        return self.subs(n - m)

    def extent(self, n1=-100, n2=100):
        """Determine extent of the signal.

        For example, nexpr([1, 1]).extent() = 2
                     nexpr([1, 0, 1]).extent() = 3
                     nexpr([0, 1, 0, 1]).extent() = 3

        This performs a search between n=n1 and n=n2."""

        return self.seq((n1, n2)).extent()

    def discrete_time_fourier_transform(self, var=None,
                                        images=oo, **assumptions):
        """Convert to Fourier domain using discrete time Fourier transform.

        Use `images = 0` to avoid the infinite number of spectral images.
        """

        return self.DTFT(var, images, **assumptions)

    def DTFT(self, var=None, images=oo, **assumptions):
        """Convert to Fourier domain using discrete time Fourier transform.

        By default this returns the DTFT in terms of `f`.  Use
        `.DTFT(w)` to get the angular frequency form, `.DTFT(F)` to
        get the normalised frequency form, or `.DTFT(W)` to get the
        normalised angular frequency form.

        Use `images = 0` to avoid the infinite number of spectral images.

        """

        from .symbols import f, omega, Omega, F
        from .fexpr import fexpr
        from .dtft import DTFT

        if var is None:
            var = f
        if id(var) not in (id(f), id(F), id(omega), id(Omega)):
            raise ValueError(
                'DTFT requires var to be f, F, omega, or Omega`, not %s' % var)

        dtft = DTFT(self.expr, self.var, fsym, images=images)

        result = fexpr(dtft)(var)
        result = result.simplify_dirac_delta()
        result = result.simplify_heaviside()
        result = result.simplify_rect()

        # There is a bug in SymPy when simplifying Sum('X(n - m)', (m, -oo, oo))
        # result = result.simplify()
        result = result.cancel_terms()

        return result

    def norm_angular_fourier(self, **assumptions):

        from .normomegaexpr import Omega
        return self.DTFT()(Omega)

    def difference_equation(self, inputsym='x', outputsym='y', form='iir'):
        """Create difference equation from impulse response.

        `form` can be 'fir' or 'iir' ('direct form I').
        """

        H = self.ZT()
        return H.difference_equation(inputsym, outputsym, form)

    def remove_condition(self):
        """Remove the piecewise condition from the expression."""

        if not self.is_conditional:
            return self
        expr = self.expr
        expr = expr.args[0].args[0]
        return self.__class__(expr)

    def force_causal(self):
        """Remove the piecewise condition from the expression
        and multiply by unit-step function.  See also remove_condition."""

        if self.is_causal:
            return self

        expr = self.expr
        if self.is_conditional:
            expr = expr.args[0].args[0]
        expr = expr * UnitStep(n.var)
        return self.__class__(expr)

    def zdomain(self, **assumptions):
        return self.ZT(**assumptions)

    def discrete_frequency(self, **assumptions):
        return self.DFT(**assumptions)

    def discrete_time(self, **assumptions):
        return self

    def fourier(self, **assumptions):
        return self.DTFT(**assumptions)

    def angular_fourier(self, **assumptions):
        from .symbols import omega

        return self.DTFT(omega, **assumptions)

    def norm_fourier(self, **assumptions):
        from .symbols import F

        return self.DTFT(F, **assumptions)


def nexpr(arg, **assumptions):
    """Create nExpr object.  If `arg` is nsym return n"""

    from .expr import Expr

    if arg is nsym:
        return n

    if isinstance(arg, Expr):
        if assumptions == {}:
            return arg
        return arg.__class__(arg, **assumptions)

    if isinstance(arg, str) and arg.startswith('{'):
        return nseq(arg)

    from numpy import ndarray

    if isinstance(arg, (list, ndarray)):
        return DiscreteTimeDomainSequence(arg, var=n).as_impulses()

    return DiscreteTimeDomainExpression(arg, **assumptions)


from .expressionclasses import expressionclasses  # nopep8

classes = expressionclasses.register(
    'discrete time', DiscreteTimeDomainExpression)
DiscreteTimeDomainVoltage = classes['voltage']
DiscreteTimeDomainCurrent = classes['current']

n = DiscreteTimeDomainExpression('n', integer=True)



"""This module provides support for nodes.

Copyright 2020--2022 Michael Hayes, UCECE

"""

from .immittancemixin import ImmittanceMixin


class Node(ImmittanceMixin):

    def __init__(self, cct, name):

        self.cct = cct
        self.name = name
        self.pos = None
        self.port = False
        parts = name.split('_')
        self.rootname = parts[0] if name[0] != '_' else name
        self.primary = len(parts) == 1
        # List of elements connected to this node.
        self._connected = []
        self._count = 0

    def __repr__(self):
        return "Node('%s')" % self.name

    @property
    def V(self):
        """Node voltage with respect to ground."""

        return self.cct.get_Vd(self.name, '0')

    @property
    def v(self):
        """Node time-domain voltage with respect to ground."""

        return self.cct.get_vd(self.name, '0')

    @property
    def admittance(self):
        """Driving-point admittance between node and ground."""

        return self.cct.admittance(self.name, '0')

    @property
    def impedance(self):
        """Driving-point impedance between node and ground."""

        return self.cct.impedance(self.name, '0')

    def append(self, cpt):

        if cpt.type in ('P', ):
            self.port = True
        if cpt.type not in ('A', 'O'):
            self._count += 1

        self._connected.append(cpt)

    @property
    def count(self):
        """Number of elements (including wires but not open-circuits and
        annotations) connected to the node"""

        return self._count

    def oneport(self, node=0):
        """Create oneport object with respect to specified node
        (default ground)."""

        return self.cct.oneport(self.name, node)

    def thevenin(self, node=0):
        """Create Thevenin oneport object with respect to specified
        node (default ground)."""

        return self.cct.thevenin(self.name, node)

    def norton(self, node=0):
        """Create Norton oneport object with respect to specified node
        (default ground)."""

        return self.cct.norton(self.name, node)

    @property
    def connected(self):
        """Return list of components connected to the node."""

        return self._connected

    def is_connected(self, cpt):
        """Return True if cpt is connected to the node."""

        if isinstance(cpt, str):
            for cpt1 in self.connected:
                if cpt1.name == cpt:
                    return True
            return False

        return cpt in self.connected

    def is_wired_to(self, node):
        """Return True if the nodes are wired together."""

        node = str(node)
        for nodes in self.cct.equipotential_nodes.values():
            if self.name in nodes and node in nodes:
                return True
        return False

    def wired_to(self):
        """Return list of names of nodes that are wired to this node."""

        for nodes in self.cct.equipotential_nodes.values():
            if self.name in nodes:
                nodes.remove(self.name)
                return nodes
        return []



"""This module defines additional functions to those defined by SymPy.
They are for internal use by Lcapy.  The user versions are defined in
function.py.

Copyright 2020--2022 Michael Hayes, UCECE

"""

import sympy as sym
from sympy.core import S, Integer
from sympy.core.logic import fuzzy_not
from .config import unitstep_zero


class Degrees(sym.Function):
    """Convert radians to degrees."""

    @classmethod
    def eval(cls, nval):
        """
        Convert radians to degrees.
        """

        return nval / sym.pi * 180


class Radians(sym.Function):
    """Convert degrees to radians."""

    @classmethod
    def eval(cls, nval):
        """
        Convert degrees to radians.
        """

        return nval / 180 * sym.pi


class UnitImpulse(sym.Function):
    """Discrete unit impulse function."""

    is_integer = True

    @classmethod
    def eval(cls, nval):
        """
        Evaluates the discrete unit impulse function.
        """

        if nval.is_zero:
            return S.One
        elif fuzzy_not(nval.is_zero):
            return S.Zero


class UnitStep(sym.Function):
    """Discrete unit step function."""

    is_integer = True

    @classmethod
    def eval(cls, nval, zero=None):
        """
        Evaluates the discrete unit step function.   This is defined
        as 1 for n >= 0 and 0 otherwise.
        """

        if nval.is_zero:
            if zero is None:
                zero = unitstep_zero
            return zero

        if nval.is_nonnegative:
            return S.One
        elif nval.is_negative:
            return S.Zero


class rect(sym.Function):
    """The rectangle function.

    rect(t) = 1 for abs(t) <= 0.5 otherwise 0."""

    @classmethod
    def eval(cls, val):
        """
        Evaluates the rectangle function.
        """

        if val.is_Number:
            if val < -0.5 or val > 0.5:
                return S.Zero
            return S.One

    def rewrite(self, *args, **hints):

        x = self.args[0]
        return sym.Heaviside(x + S.Half) - sym.Heaviside(x - S.Half)


class dtrect(sym.Function):
    """The discrete-time rectangle function."""

    @classmethod
    def eval(cls, val):
        """
        Evaluates the rectangle function for discrete-time signals.
        """

        if val.is_Number:
            if val < -0.5 or val >= 0.5:
                return S.Zero
            return S.One

    def rewrite(self, *args, **hints):

        x = self.args[0]
        return UnitStep(x + S.Half) - UnitStep(x - S.Half)


class dtsign(sym.Function):
    """The discrete-time signum function."""

    @classmethod
    def eval(cls, val):
        """
        Evaluates the signum function for discrete-time signals.
        """

        if val.is_Number:
            if val >= 0:
                return S.One
            return S.Zero


class sincn(sym.Function):
    """Normalized sinc function :math:`\sin(\pi x)/(\pi x)`."""

    @classmethod
    def eval(cls, val):
        """
        Evaluates the normalized sinc (cardinal sine) function.
        This is what NumPy uses but not SymPy.
        """

        if val.is_Number:
            if val == 0:
                return S.One
            x = sym.pi * val
            return sym.sin(x) / x

    def rewrite(self, *args, **hints):

        x = sym.pi * self.args[0]
        return sym.sin(x) / x


class sincu(sym.Function):
    """Unnormalized sinc function :math:`\sin(x)/(x)`."""

    @classmethod
    def eval(cls, val):
        """
        Evaluates the unnormalized sinc (cardinal sine) function.
        This is what SymPy uses but not NumPy.
        """

        if val.is_Number:
            if val == 0:
                return S.One
            return sym.sin(val) / val

    def rewrite(self, *args, **hints):

        x = self.args[0]
        return sym.sin(x) / x


class psinc(sym.Function):
    """Periodic sinc function :math:`\sin(M * pi * x)/(M * sin(pi * x))`."""

    @classmethod
    def eval(cls, M, val):
        """
        Evaluates the periodic sinc function.
        """
        if val.is_Number:
            if val == 0:
                return S.One
            if val.is_integer and val.is_even:
                return S.One
            if val.is_integer and val.is_odd:
                return -S.One

            x = sym.pi * val
            return sym.sin(M * x) / (M * sym.sin(x))

    def rewrite(self, *args, **hints):

        M = self.args[0]
        x = sym.pi * self.args[1]
        # If evaluate, SymPy will convert sin(n * pi) to 0.
        return sym.sin(M * x, evaluate=False) / (M * sym.sin(x, evaluate=False))


class tri(sym.Function):
    """The triangle function.

    tri(t) = 1 - abs(t) for abs(t) <= 1 otherwise 0."""

    @classmethod
    def eval(cls, val):
        """
        Evaluates the triangle function.
        """

        if val.is_Number:

            if val >= 1:
                return S.Zero
            elif val <= -1:
                return S.Zero
            else:
                return 1 - abs(val)

    def rewrite(self, *args, **hints):

        x = self.args[0]
        return ramp(x + 1) - 2 * ramp(x) + ramp(x - 1)


class trap(sym.Function):
    """The trapezoidal function."""

    @classmethod
    def eval(cls, val, alpha):
        """
        Evaluates the trapezoid function.   This is rect(t / alpha).convolve(rect(t)).
        trap(t, 0) = rect(t) and trap(t, 1) = tri(t).
        """

        if val.is_Number and alpha.is_number:
            absval = abs(val)
            foo = absval - 0.5

            if alpha == S.Zero:
                if val < -0.5 or val > 0.5:
                    return S.Zero
                return S.One

            if foo >= 0.5 * alpha:
                return S.Zero
            elif foo <= -0.5 * alpha:
                return S.One
            else:
                return 0.5 - foo / alpha

    def rewrite(self, *args, **hints):

        x = self.args[0]
        alpha = self.args[1]
        if alpha == 0:
            return rect(x)
        elif alpha == 1:
            return tri(x)
        return rampstep((x + (1 + alpha) / 2) / alpha) - rampstep((x - (1 - alpha) / 2) / alpha)


class ramp(sym.Function):
    """The ramp function.

    ramp(t) = t for t > 0 otherwise 0."""

    @classmethod
    def eval(cls, val):
        """
        Evaluates the ramp function.
        """

        if val.is_Number:

            if val >= 0:
                return val
            else:
                return 0

    def rewrite(self, *args, **hints):

        x = self.args[0]
        return x * sym.Heaviside(x)


class rampstep(sym.Function):
    """The rampstep function.

    rampstep(t) = t for 0 < t < 1, 1 for t > 1 and otherwise 0."""

    @classmethod
    def eval(cls, val):
        """
        Evaluates the rampstep function.
        """

        if val.is_Number:

            if val >= 0 and val < 1:
                return val
            elif val > 1:
                return 1
            else:
                return 0

    def rewrite(self, *args, **hints):

        x = self.args[0]
        return ramp(x) - ramp(x - 1)



"""This module provides continuous-time filter support.

Copyright 2022 Michael Hayes, UCECE

"""

from .expr import expr, equation, ExprTuple
from .differentialequation import DifferentialEquation
from .functions import Derivative
from .texpr import TimeDomainExpression
from .symbols import t, s
from .utils import isiterable
import sympy as sym


class LTIFilter(object):

    def __init__(self, b, a):
        """Create continuous-time filter where `b` is a list or array of
        numerator coefficients and `a` is a list or array of
        denominator coefficients."""

        if not isiterable(b):
            b = (b, )
        if not isiterable(a):
            a = (a, )

        self.a = ExprTuple(a)
        self.b = ExprTuple(b)

    def __repr__(self):
        """This is called by repr(expr).  It is used, e.g., when printing
        in the debugger."""

        return '%s(%s, %s)' % (self.__class__.__name__, self.b, self.a)

    def transfer_function(self):
        """Return continuous-time impulse response (transfer function) in
        s-domain."""

        from .sym import ssym

        a = self.a.sympy
        b = self.b.sympy

        Nl = len(a)
        Nr = len(b)

        # Numerator of H(s)
        num = sym.S.Zero
        for i in range(Nr):
            num += b[i] * ssym**(Nr - i - 1)

        # Denominator for H(s)
        denom = sym.S.Zero
        for k in range(Nl):
            denom += a[k] * ssym**(Nl - k - 1)

        # Collect with respect to positive powers of the variable s
        num = sym.collect(sym.expand(num), ssym)
        denom = sym.collect(sym.expand(denom), ssym)

        Hs = expr(sym.simplify(num / denom))
        Hs.is_causal = True
        return Hs

    def impulse_response(self):
        """Return continuous-time impulse response (transfer function) in time
        domain."""

        H = self.transfer_function()
        return H(t)

    def step_response(self):
        """Return continuous-time step response in time domain."""

        H = self.transfer_function()
        G = H / s
        G.is_causal = H.is_causal

        return G(t)

    def frequency_response(self, **assumptions):
        """Return frequency response."""

        return self.transfer_function().FT(**assumptions)

    def differential_equation(self, inputsym='x', outputsym='y'):
        """Return differential equation."""

        rhs = 0 * t

        for m, bn in enumerate(self.b):
            rhs += bn * Derivative(expr('%s(t)' % inputsym), t, m)

        for m, an in enumerate(self.a[1:]):
            rhs -= an * Derivative(expr('%s(t)' % outputsym), t, m + 1)

        lhs = self.a[0] * expr('y(t)')
        return DifferentialEquation(lhs, rhs, inputsym, outputsym)

    def sdomain_initial_response(self, ic):
        """Return s-domain response due to initial conditions.
           ic : list of initial conditions y(0), y'(0), y''(0), ...

        """

        from .sym import ssym

        Nl = len(self.a)
        if len(ic) != Nl:
            raise ValueError('Expecting %d initial conditions, got %d'
                             % (Nl, len(ic)))

        # Denominator for Yi(s)
        denom = self.a[0] * s**0
        num = 0 * s
        for k in range(1, Nl):
            aas = self.a[k] * s**(-k)
            denom += aas
            # Numerator for Yi(s)
            y0 = 0 * s
            for i in range(0, k):
                y0 += ic[i] * s**(i + 1)
            num += aas * y0

        # Collect with respect to positive powers of the variable s
        num = num.sympy * ssym**Nl
        denom = denom.sympy * ssym**Nl

        num = sym.collect(sym.expand(num), ssym)
        denom = sym.collect(sym.expand(denom), ssym)

        Ysi = expr(-sym.simplify(num / denom))
        Ysi.is_causal = True

        return Ysi

    def initial_response(self, ic):
        """Return response due to initial conditions in the time domain.
           ic : list of initial conditions y(0), y'(0), y''(0), ...
        """

        Ysi = self.sdomain_initial_response(ic)
        return Ysi(t)

    def response(self, x, ic=None, ni=None):
        """Calculate response of filter to input `x` given a list of initial conditions
        `ic` for time indexes specified by `ni`.  If `ni` is a tuple,
        this specifies the first and last (inclusive) time index.

        The initial conditions are valid prior to the time indices given by the ni
        `x` can be an expression, a sequence, or a list/array of values.
        """

        return 0

    def subs(self, *args, **kwargs):

        a = self.a.subs(*args, **kwargs)
        b = self.b.subs(*args, **kwargs)

        return self.__class__(b, a)

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self

    @property
    def is_stable(self):
        """Return True if all the poles of the filter's transfer function
        are within the unit circle."""

        return self.transfer_function().is_stable



"""This module provides the LaplaceDomainExpression class to represent
s-domain (Laplace domain) expressions.

Copyright 2014--2022 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import LaplaceDomain
from .inverse_laplace import inverse_laplace_transform
from .state import state
from .sym import ssym, tsym, j, pi, sympify
from .ratfun import _zp2tf, _pr2tf, Ratfun
from .expr import Expr, symbol, expr, ExprDict, ExprList, exprcontainer, expr_make
from .differentialequation import DifferentialEquation
from .units import u as uu
from .functions import sqrt, DiracDelta
from .sym import dt
from sympy import limit, exp, Poly, Derivative, Integral, div, oo, Eq, Expr as symExpr
from warnings import warn


__all__ = ('sexpr', 'zp2tf', 'tf', 'pr2tf')


class LaplaceDomainExpression(LaplaceDomain, Expr):
    """s-domain expression or symbol."""

    var = ssym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)
        super(LaplaceDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr
        # Be more specific
        if check and expr.has(tsym) and not expr.has(Derivative) and not expr.has(Integral):
            raise ValueError(
                's-domain expression %s cannot depend on t' % expr)

    def as_expr(self):
        return LaplaceDomainExpression(self)

    @classmethod
    def from_poles_residues(cls, poles, residues):
        """Create a transfer function from lists of poles and residues.

        See also from_zeros_poles_gain, from_numer_denom"""

        return cls(pr2tf(poles, residues, cls.var), causal=True)

    @classmethod
    def from_zeros_poles_gain(cls, zeros, poles, K=1):
        """Create a transfer function from lists of zeros and poles,
        and from a constant gain.

        See also from_poles_residues, from_numer_denom"""

        return cls(zp2tf(zeros, poles, K, cls.var), causal=True)

    @classmethod
    def from_numer_denom(cls, numer, denom):
        """Create a transfer function from lists of the coefficient
        for the numerator and denominator.

        See also from_zeros_poles_gain, from_poles_residues"""

        return cls(tf(numer, denom, cls.var), causal=True)

    def tdifferentiate(self):
        """Differentiate in t-domain (multiply by s)."""

        return self.__class__(self.expr * self.var, **self.assumptions)

    def tintegrate(self):
        """Integrate in t-domain (divide by s)."""

        return self.__class__(self.expr / self.var, **self.assumptions)

    def delay(self, T):
        """Apply delay of T seconds by multiplying by exp(-s T)."""

        T = self.__class__(T)
        return self.__class__(self.expr * exp(-s * T))

    @property
    def jomega(self):
        """Return expression with s = j omega."""

        from .symbols import jomega
        return self.subs(self.var, jomega)

    def post_initial_value(self):
        """Determine post-initial value at t = 0+."""

        return self.__class__(limit(self.expr * self.var, self.var, oo))

    def final_value(self):
        """Determine value at t = oo."""

        return self.__class__(limit(self.expr * self.var, self.var, 0))

    def inverse_laplace(self, zero_initial_conditions=None, **assumptions):
        """Attempt inverse Laplace transform.

        If causal=True the response is zero for t < 0 and
        the result is multiplied by Heaviside(t)
        If ac=True or dc=True the result is extrapolated for t < 0.
        Otherwise the result is only known for t >= 0.

        By default initial conditions are assumed to be zero.  This
        can be controlled by `zero_initial_conditions`.

        """

        return self.ILT(zero_initial_conditions=zero_initial_conditions, **assumptions)

    def ILT(self, zero_initial_conditions=None, **assumptions):
        """Attempt inverse Laplace transform.

        If causal=True the response is zero for t < 0 and
        the result is multiplied by Heaviside(t)
        If ac=True or dc=True the result is extrapolated for t < 0.
        Otherwise the result is only known for t >= 0.

        By default initial conditions are assumed to be zero.  This
        can be controlled by `zero_initial_conditions`.

        """
        if zero_initial_conditions is None:
            zero_initial_conditions = state.zero_initial_conditions

        assumptions = self.assumptions.merge(**assumptions)
        result = inverse_laplace_transform(self.expr, self.var, tsym,
                                           zero_initial_conditions=zero_initial_conditions,
                                           **assumptions)

        return self.change(result, domain='time', units_scale=uu.Hz,
                           **assumptions)

    def time(self, **assumptions):
        """Convert to time domain.

        If causal=True the response is zero for t < 0 and
        the result is multiplied by Heaviside(t)
        If ac=True or dc=True the result is extrapolated for t < 0.
        Otherwise the result is only known for t >= 0.

        """

        try:
            return self.inverse_laplace(**assumptions)
        except ValueError:
            return self.as_sum().inverse_laplace(**assumptions)

    def laplace(self, **assumptions):
        """Convert to s-domain."""

        assumptions = self.assumptions.merge(**assumptions)
        return self.__class__(self, **assumptions)

    def fourier(self, **assumptions):
        """Convert to Fourier domain."""
        from .symbols import f, jw, pi

        if self.is_causal or assumptions.get('causal', False):
            # Note, this does not apply for 1 / s.
            tmp = self.subs(jw)
            if tmp.real != 0:
                return self.change(tmp.subs(2 * pi * f, safe=True),
                                   domain='fourier', **assumptions)

        result = self.time(**assumptions).fourier(**assumptions)
        return result

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""
        from .symbols import jw

        if self.is_causal:
            # Note, this does not apply for 1 / s.
            tmp = self(jw)
            if tmp.real != 0:
                return self.change(tmp, domain='angular fourier',
                                   **assumptions)

        result = self.time(**assumptions).angular_fourier(**assumptions)
        return result

    def norm_angular_fourier(self, **assumptions):
        """Convert to normalized angular Fourier domain."""
        from .symbols import jw, Omega

        if self.is_causal:
            # Note, this does not apply for 1 / s.
            tmp = self(j * Omega / dt)
            if tmp.real != 0:
                return self.change(tmp, domain='norm angular fourier',
                                   **assumptions)

        result = self.time(**assumptions).norm_angular_fourier(**assumptions)
        return result

    def norm_fourier(self, **assumptions):
        """Convert to normalized Fourier domain."""
        from .symbols import jw, F

        if self.is_causal:
            # Note, this does not apply for 1 / s.
            tmp = self(j * F / dt)
            if tmp.real != 0:
                return self.change(tmp, domain='norm fourier',
                                   **assumptions)

        result = self.time(**assumptions).norm_fourier(**assumptions)
        return result

    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        result = PhasorFrequencyDomainExpression.from_laplace(
            self, **assumptions)
        return result

    def transient_response(self, tvector=None):
        """Evaluate transient (impulse) response."""

        if tvector is None:
            return self.time()

        return self.time().evaluate(tvector)

    def impulse_response(self, tvector=None):
        """Evaluate transient (impulse) response."""

        return self.transient_response(tvector)

    def step_response(self, tvector=None):
        """Evaluate step response."""

        H = self.__class__(self / self.var, **self.assumptions)
        return H.transient_response(tvector)

    def angular_frequency_response(self, wvector=None):
        """Convert to angular frequency domain and evaluate response if
        angular frequency vector specified.

        """
        from .symbols import omega

        X = self.subs(j * omega)

        if wvector is None:
            return X

        return X.evaluate(wvector)

    def frequency_response(self, fvector=None):
        """Convert to frequency domain and evaluate response if frequency
        vector specified.

        """
        from .symbols import f

        X = self.subs(j * 2 * pi * f)

        if fvector is None:
            return X

        return X.evaluate(fvector)

    def _response_impulse_invariance(self, xvector, tvector, dtval):
        """Evaluate response of system with applied signal.
        This method assumes that the data is well over-sampled."""

        from numpy import arange, convolve, diff, hstack

        # Perform polynomial long division so expr1 = Q + M / A
        B, A, delay, undef = self._as_B_A_delay_undef()
        if undef != 1:
            raise ValueError('Have undefined expression %s' % undef)
        Q, M = div(B, A, self.var)
        expr1 = M / A

        Nt = len(tvector)

        # Evaluate impulse response.
        th = arange(Nt) * dtval
        H = LaplaceDomainExpression(expr1, **self.assumptions)
        hvector = H.transient_response(th)

        ty = tvector
        y = convolve(xvector, hvector)[0:Nt] * dtval

        if Q:
            # Handle Dirac deltas and their derivatives.
            C = expr(Q).coeffs()
            for c in reversed(C):

                y += float(c.expr) * xvector

                xvector = diff(xvector) / dtval
                xvector = hstack((xvector, 0))

        from scipy.interpolate import interp1d

        if delay != 0.0:
            # Try linear interpolation; should oversample first...
            y = interp1d(ty, y, bounds_error=False, fill_value=0)
            y = y(tvector - float(delay))

        return y

    def _response_bilinear(self, xvector, tvector, dtval, alpha=0.5):

        import scipy.signal as signal
        from numpy import hstack, zeros

        expr, delay = self.as_ratfun_delay()
        Ndelay = 0
        if delay != 0:
            if delay < 0:
                raise ValueError('Need time advance %s' % delay)
            Ndelay = round(float(delay / dtval), 12)
            if not Ndelay.is_integer:
                # Require fractional delay.
                Ndelay = 0
                delay = 0
                expr = self
            else:
                Ndelay = int(Ndelay)
        if delay != 0:
            xvector = xvector[:-Ndelay]
            tvector = tvector[:-Ndelay] - delay

        if expr.has(exp):
            warn('Using second order Pade approximation for exp.')
            expr = self.approximate_exp(method='pade', order=2, numer_order=1)

        Hz = expr.generalized_bilinear_transform(alpha).subs(dt, dtval)
        fil = Hz.dlti_filter()

        a = [a1.fval for a1 in fil.a]
        b = [b1.fval for b1 in fil.b]

        y = signal.lfilter(b, a, xvector)

        if Ndelay != 0:
            y = hstack((zeros(Ndelay), y))
        return y

    def response(self, xvector, tvector, method='bilinear', alpha=0.5):
        """Evaluate response to input signal `xvector` at times
        specified by `tvector`.  This returns a NumPy array.

        The default method is 'bilinear'.  Other methods are:
        'impulse-invariance'
        'bilinear', 'tustin', 'trapezoidal'
        'generalized-bilinear', 'gbf' controlled by the parameter `alpha`
        'euler', 'forward-diff', 'forward-euler'
        'backward-diff', 'backward-euler'
        """

        from numpy import array, diff, allclose

        xvector = array(xvector)
        tvector = array(tvector)

        symbols = self.symbols
        symbols.pop('s', None)
        if symbols != {}:
            raise ValueError('Have undefined symbols: %s' % symbols)

        if len(xvector) != len(tvector):
            raise ValueError('xvector, tvector must have at least 2 samples')

        if len(xvector) < 2:
            raise ValueError('xvector must have same length as tvector')

        td = diff(tvector)
        if not allclose(diff(td), 0):
            raise ValueError('Time samples not uniformly spaced')

        dtval = td[0]
        if dtval < 0:
            raise ValueError('Time samples not increasing')

        if self.is_constant:
            return float(self.expr) * xvector

        # Expand cosh, sinh into sum of exps.
        expr = self.expand_hyperbolic_trig()

        if method in ('impulse-invariance', 'adhoc'):
            result = expr._response_impulse_invariance(xvector, tvector, dtval)
        elif method in ('bilinear', 'tustin', 'trapezoidal'):
            result = expr._response_bilinear(
                xvector, tvector, dtval, alpha=0.5)
        elif method in ('gbf', 'generalized-bilinear'):
            result = expr._response_bilinear(
                xvector, tvector, dtval, alpha=alpha)
        elif method in ('euler', 'forward-diff', 'forward-euler'):
            result = expr._response_bilinear(xvector, tvector, dtval, alpha=0)
        elif method in ('backward-diff', 'backward-euler'):
            result = expr._response_bilinear(xvector, tvector, dtval, alpha=1)
        else:
            raise ValueError('Unknown method %s' % method)

        return result * dtval

    def state_space(self, form='CCF'):
        """Create state-space representation from transfer function.  Note,
        state-space representations are not unique and are determined
        by the `form` argument.  Currently this can be 'CCF' for the
        controllable canonical form, 'OCF' for the observable
        canonical form, or 'DCF' for the diagonal canonical form."""

        from .statespace import StateSpace

        a = self.a
        b = self.b

        return StateSpace.from_transfer_function_coeffs(b, a, form)

    @property
    def ss(self):
        """Return state-space representation using controllable canonical form.
        For other forms, use `state_space()`."""

        return self.state_space()

    def differential_equation(self, inputsym='x', outputsym='y'):
        """Create differential equation from transfer function.

        For example,
        >>> H = (s + 3) / (s**2 + 4)
        >>> H.differential_equation()
                 d                    d
        3.y(t) + --(y(t)) = 4.x(t) + ---(x(t))
                 dt                    2
                                     dt
        """

        x = texpr('%s(t)' % inputsym)
        y = texpr('%s(t)' % outputsym)

        X = x.LT()
        Y = y.LT()

        H = self.simplify()
        N = H.N
        D = H.D

        lhs = (N * Y).ILT(causal=True, zero_initial_conditions=True)
        rhs = (D * X).ILT(causal=True, zero_initial_conditions=True)

        return DifferentialEquation(lhs, rhs, inputsym, outputsym)

    def lti_filter(self, normalize_a0=True):
        """Create continuous-time linear time-invariant filter from
        continuous-time transfer function."""

        # TODO, perhaps add only to TimeDomainTransfer
        # or check that is_transfer, is_impedance, or is_admittance

        from .ltifilter import LTIFilter

        if not self.is_rational_function:
            raise ValueError("Not a rational function")

        N = self.N
        D = self.D
        bn = N.coeffs()
        an = D.coeffs()

        if normalize_a0:
            bn = [bx / an[0] for bx in bn]
            an = [ax / an[0] for ax in an]

        lpf = LTIFilter(bn, an)
        return lpf

    def dlti_filter(self, method='bilinear'):
        """Create DLTI filter using bilinear transform."""

        if method != 'bilinear':
            raise ValueError('Unsupported transform ' + method)
        return self.bilinear_transform().simplify().dlti_filter()

    def evaluate(self, svector=None):

        return super(LaplaceDomainExpression, self).evaluate(svector)

    def plot(self, **kwargs):
        """Plot pole-zero map.

        kwargs include:
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label (default Re(s))
        ylabel - the y-axis label (default Im(s))
        xscale - the x-axis scaling
        yscale - the y-axis scaling
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned."""

        from .plot import plot_pole_zero
        return plot_pole_zero(self, **kwargs)

    def pole_zero_plot(self, **kwargs):
        """Plot pole-zero map."""

        return self.plot(**kwargs)

    def bode_plot(self, fvector=None, unwrap=True, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Bode
        plot (but without the straight line approximations).  fvector
        specifies the frequencies.  If it is a tuple (f1, f2), it sets
        the frequency limits.   Since a logarithmic frequency scale is used,
        f1 must be greater than 0.

        `unwrap` controls phase unwrapping (default True).

        This method makes the assumption that the expression is causal.
        Note, the Bode plot does not exist for marginally stable and unstable
        systems since `jw` is outside the region of convergence.
        """

        return self.fourier(causal=True).bode_plot(fvector, unwrap=unwrap,
                                                   **kwargs)

    def nyquist_plot(self, fvector=None, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Nyquist
        plot (imaginary part versus real part).  fvector specifies the
        frequencies.  If it is a tuple (f1, f2), it sets the frequency
        limits.

        `npoints` set the number of plotted points.

        The unit circle is shown by default.  This can be disabled with `unitcircle=False`.

        This method makes the assumption that the expression is causal.
        """

        return self.fourier(causal=True).nyquist_plot(fvector, **kwargs)

    def nichols_plot(self, fvector=None, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Nichols
        plot (dB versus phase).  fvector specifies the frequencies.
        If it is a tuple (f1, f2), it sets the frequency limits.

        `npoints` set the number of plotted points.

        This method makes the assumption that the expression is causal.

        """

        return self.fourier(causal=True).nichols_plot(fvector, **kwargs)

    def generalized_bilinear_transform(self, alpha=0.5):

        from .discretetime import z, dt

        if alpha < 0 or alpha > 1:
            raise ValueError("alpha must be between 0 and 1 inclusive")

        return self.subs((1 / dt) * (1 - z**-1) / (alpha + (1 - alpha) * z**-1)) / dt

    def bilinear_transform(self):
        """Approximate s = ln(z) / dt

        by s = (2 / dt) * (1 - z**-1) / (1 + z**-1)

        This is also called Tustin's method and is equivalent to the
        trapezoidal method."""

        # TODO: add frequency warping as an option
        return self.generalized_bilinear_transform(0.5)

    def forward_euler_transform(self):
        """Approximate s = ln(z)

        by s = (1 / dt) * (1 - z**-1) / z**-1.   This is also known
        as the forward difference method."""

        return self.generalized_bilinear_transform(0)

    def backward_euler_transform(self):
        """Approximate s = ln(z)

        by s = (1 / dt) * (1 - z**-1).  This is also known
        as the backward difference method."""

        return self.generalized_bilinear_transform(1)

    def simpson_transform(self):
        """Approximate s = ln(z)

        by s = (3 / dt) * (z**2 - 1) / (z**2 + 4 * z + 1).  This
        doubles the system order and can produce unstable poles."""

        from .discretetime import z

        return self.subs((3 / dt) * (z**2 - 1) / (z**2 + 4 * z + 1))

    def matched_ztransform(self):
        """Match poles and zeros of H(s) to approximate H(z).

        If there are no zeros, this is equivalent to impulse_invariance.

        See also bilinear_transform and impulse_invariance_transform."""

        from .discretetime import z

        zeros, poles, K, undef = self._ratfun.as_ZPK()
        result = K
        for zero in zeros:
            result *= (1 - exp(zero * dt) / z)
        for pole in poles:
            result /= (1 - exp(pole * dt) / z)
        result *= undef

        result.is_causal = self.is_causal
        return result

    def impulse_invariance_transform(self):
        """This samples the impulse response and then calculates the
        Z-transform.  It does not work if the impulse response
        has Dirac deltas, say for a transfer function that is
        a pure delay or is not-strictly proper.

        The discrete-time and continuous-time impulse responses
        are identical at the sampling instants n * dt.

        The data needs to be sampled many times the bandwidth to avoid
        aliasing.

        See also bilinear_transform and matched_ztransform."""

        from .discretetime import n, z, dt

        # An alternative approach is to expand as partial fractions
        # and then replace (s + alpha) with (1 - exp(-alpha * dt) *
        # z**-1) in the denominator of each partial fraction (the
        # residues are unchanged).  This maps the pole at -alpha to
        # exp(-alpha * dt).

        h = self.ILT()
        if h.has(DiracDelta):
            raise ValueError('Impulse response has Dirac-deltas')

        hn = h.subs(n * dt)
        H = hn.ZT()
        return H

    def discretize(self, method='bilinear', alpha=0.5):
        """Convert to a discrete-time approximation in the z-domain:

        :math:`H(z) \approx H_c(s)`

        If :math:`H(s)` is a transfer function then for the
        impulse-invariance method, the discrete-time impulse response
        is related to the continuous-time impulse response by

        :math:`h[n] = h_c(n \Delta t)`

        Note, when designing digital filters, it is often common to to
        scale the discrete-time impulse response by the sampling
        interval:

        :math:`h[n] = \Delta t h_c(n \Delta t)`

        The default method is 'bilinear'.  Other methods are:
        'impulse-invariance' 'bilinear', 'tustin', 'trapezoidal'
        'generalized-bilinear', 'gbf' controlled by the parameter
        `alpha` 'euler', 'forward-diff', 'forward-euler'
        'backward-diff', 'backward-euler' 'simpson', 'matched-Z',
        'zero-pole-matching'

        """

        if method in ('gbf', 'generalized-bilinear'):
            return self.generalized_bilinear_transform(alpha)
        elif method in ('bilinear', 'tustin', 'trapezoidal'):
            return self.generalized_bilinear_transform(0.5)
        elif method in ('euler', 'forward-diff', 'forward-euler'):
            return self.generalized_bilinear_transform(0)
        elif method in ('backward-diff', 'backward-euler'):
            return self.generalized_bilinear_transform(1)
        elif method in ('simpson', ):
            return self.simpson_transform()
        elif method in ('impulse-invariance', ):
            return self.impulse_invariance_transform()
        elif method in ('matched-Z', 'zero-pole-matching'):
            return self.matched_ztransform()
        else:
            raise ValueError('Unsupported method %s' % method)

    def zdomain(self, **assumptions):
        return self.discretize(**assumptions)

    def discrete_frequency(self, method='bilinear', **assumptions):
        return self.zdomain(method=method).discrete_frequency(**assumptions)

    def discrete_time(self, method='bilinear', **assumptions):
        return self.zdomain(method=method).discrete_time(**assumptions)


def tf(numer, denom=1, var=None):
    """Create a transfer function from lists of the coefficient
    for the numerator and denominator."""

    if var is None:
        var = ssym

    N = Poly(sympify(numer), var)
    D = Poly(sympify(denom), var)

    return LaplaceDomainTransferFunction(N / D, causal=True)


def zp2tf(zeros, poles, K=1, var=None):
    """Create a transfer function from lists (or dictionaries) of zeros and poles,
    and from a constant gain."""

    if var is None:
        var = ssym
    return LaplaceDomainTransferFunction(_zp2tf(sympify(zeros), sympify(poles),
                                                sympify(K), var), causal=True)


def pr2tf(poles, residues, var=None):
    """Create a transfer function from lists of poles and residues."""

    if var is None:
        var = ssym
    return LaplaceDomainTransferFunction(_pr2tf(sympify(poles),
                                                sympify(residues), var),
                                         causal=True)


def sexpr(arg, **assumptions):
    """Create LaplaceDomainExpression object.  If `arg` is ssym return s"""

    if arg is ssym:
        return s
    return expr_make('laplace', arg, **assumptions)


from .expressionclasses import expressionclasses  # nopep8

classes = expressionclasses.register('laplace', LaplaceDomainExpression)
LaplaceDomainVoltage = classes['voltage']
LaplaceDomainCurrent = classes['current']
LaplaceDomainAdmittance = classes['admittance']
LaplaceDomainImpedance = classes['impedance']
LaplaceDomainTransferFunction = classes['transfer']

from .texpr import TimeDomainExpression, texpr  # nopep8
from .phasor import PhasorFrequencyDomainExpression  # nopep8

s = LaplaceDomainExpression('s')
s.units = uu.rad / uu.s



from lcapy import *
import numpy as np
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_sim(self):

        cct = Circuit("""
        V1 1 0 step 10; down
        R1 1 2 5; right
        L1 2 0_2 2; down
        W 0 0_2; right""")

        results = cct.sim(np.linspace(0, 10))

        self.assertEqual(results.R1.v[0], 0, 'vR[0]')
        self.assertEqual(np.round(results.R1.v[-1], 3), 10, 'vR[-1]')
        self.assertEqual(np.round(results.R1.i[-1], 3), 2, 'iR[-1]')
        self.assertEqual(np.round(results.L1.v[-1], 3), 0, 'vL[0]')
        self.assertEqual(np.round(results.L1.i[-1], 3), 2, 'iL[-1]')



import sympy as sym
from sympy.printing.str import StrPrinter
from sympy.printing.latex import LatexPrinter
from sympy.printing.pretty.pretty import PrettyPrinter


class PrintingConfig(object):

    def __init__(self):

        self._abbreviate_units = False
        self._order = None

    @property
    def abbreviate_units(self):

        return self._abbreviate_units

    @abbreviate_units.setter
    def abbreviate_units(self, val):

        self._abbreviate_units = val

        # Print abbreviated units, V not volt
        StrPrinter._default_settings['abbrev'] = val

    @property
    def order(self):

        return self._order

    @order.setter
    def order(self, val):
        """Use 'none' to disable symbol ordering when expressions are printed."""

        self._order = val

        StrPrinter._default_settings['order'] = val
        LatexPrinter._default_settings['order'] = val
        PrettyPrinter._default_settings['order'] = val



"""This module provides the ZDomainExpression class to represent z-domain expressions.

Copyright 2020--2022 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import ZDomain
from .inverse_ztransform import inverse_ztransform
from .sym import j, pi, fsym, omegasym
from .sym import nsym, ksym, zsym, dt
from .vector import Vector
from .ratfun import _zp2tf, Ratfun
from .expr import symbol, expr, ExprDict, ExprList
from .differenceequation import DifferenceEquation
from .seqexpr import SequenceExpression
from .zseq import ZDomainSequence
from .functions import sqrt, exp
from sympy import Eq, div, limit, oo, Sum


__all__ = ('zexpr', )


class ZDomainExpression(ZDomain, SequenceExpression):
    """z-domain expression or symbol."""

    var = zsym
    seqcls = ZDomainSequence

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)

        super(ZDomainExpression, self).__init__(val, **assumptions)

        expr = self.expr
        if check and expr.has(nsym) and not expr.has(Sum):
            raise ValueError(
                'z-domain expression %s cannot depend on n' % expr)
        if check and expr.has(ksym) and not expr.has(Sum):
            raise ValueError(
                'z-domain expression %s cannot depend on k' % expr)

    def as_expr(self):
        return ZDomainExpression(self)

    def ndifferentiate(self):
        """First order difference in n-domain."""

        q = 1 / (1 - 1 / self.var) * dt
        return self.__class__(self.expr / q, **self.assumptions)

    def nintegrate(self):
        """First order integration in n-domain."""

        q = 1 / (1 - 1 / self.var) * dt
        return self.__class__(self.expr * q, **self.assumptions)

    def initial_value(self):
        """Determine value at n = 0."""

        return self.__class__(limit(self.expr * self.var, self.var, oo))

    def final_value(self):
        """Determine value at n = oo."""

        return self.__class__(limit(self.expr * self.var, self.var, 0))

    def inverse_ztransform(self, **assumptions):
        """Attempt inverse Z ransform.

        If causal=True the response is zero for n < 0 and
        the result is multiplied by UnitStep(n)
        If ac=True or dc=True the result is extrapolated for n < 0.
        Otherwise the result is only known for n >= 0.

        """

        assumptions = self.assumptions.merge(**assumptions)
        result = inverse_ztransform(self.expr, self.var, nsym, **assumptions)
        return self.change(result, domain='discrete time', **assumptions)

    def IZT(self, **assumptions):
        return self.inverse_ztransform(**assumptions)

    def transient_response(self, tvector=None):
        """Evaluate transient (impulse) response."""

        if tvector is None:
            return self.time()

        return self.time().evaluate(tvector)

    def impulse_response(self, tvector=None):
        """Evaluate transient (impulse) response."""

        return self.transient_response(tvector)

    def step_response(self, tvector=None):
        """Evaluate step response."""

        q = 1 / (1 - 1 / self.var)
        H = self.__class__(self * q, **self.assumptions)
        return H.transient_response(tvector)

    def frequency_response(self, fvector=None, var=None):
        """Convert to frequency domain using DTFT and evaluate response if
        frequency vector specified."""

        X = self.DTFT(var)

        if fvector is None:
            return X

        return X.evaluate(fvector)

    def response(self, x, t):
        """Evaluate response to input signal x at times t."""

        from numpy import allclose, diff, ones, zeros, arange, convolve, hstack

        if len(x) != len(t):
            raise ValueError('x must have same length as t')

        dt = t[1] - t[0]
        if not allclose(diff(t), ones(len(t) - 1) * dt):
            raise (ValueError, 't values not equally spaced')

        # Perform polynomial long division so expr = Q + M / D
        N, D, delay, undef = self._as_N_D_delay_undef()
        if undef != 1:
            raise ValueError('Have undefined expression %s' % undef)
        Q, M = div(N, D)
        expr = M / D

        N = len(t)

        # Evaluate transient response.
        th = arange(N) * dt - dt
        h = ZDomainExpression(expr).transient_response(th)

        print('Convolving...')
        ty = t
        y = convolve(x, h)[0:N] * dt

        if Q:
            # Handle Dirac deltas and their derivatives.
            C = Q.all_coeffs()
            for n, c in enumerate(C):

                y += c * x

                x = diff(x) / dt
                x = hstack((x, 0))

        from scipy.interpolate import interp1d

        if delay != 0.0:
            print('Interpolating...')
            # Try linear interpolation; should oversample first...
            y = interp1d(ty, y, bounds_error=False, fill_value=0)
            y = y(t - delay)

        return y

    def state_space(self, form='CCF'):
        """Create state-space representation from transfer function.  Note,
        state-space representations are not unique and are determined
        by the `form` argument.  Currently this can be 'CCF' for the
        controllable canonical form, 'OCF' for the observable
        canonical form, or 'DCF' for the diagonal canonical form."""

        from .dtstatespace import DTStateSpace

        a = self.a
        b = self.b

        return DTStateSpace.from_transfer_function_coeffs(b, a, form)

    @property
    def ss(self):
        """Return state-space representation using controllable canonical form.
        For other forms, use `state_space()`."""

        return self.state_space()

    def evaluate(self, svector=None):

        return super(ZDomainExpression, self).evaluate(svector)

    def plot(self, t=None, **kwargs):
        """Plot pole-zero map.

        kwargs include:
        unitcircle - if True, draw unit circle
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label (default Re(z))
        ylabel - the y-axis label (default Im(z))
        xscale - the x-axis scaling
        yscale - the y-axis scaling
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned."""

        if 'unitcircle' not in kwargs:
            kwargs['unitcircle'] = True

        from .plot import plot_pole_zero
        return plot_pole_zero(self, **kwargs)

    def bode_plot(self, fvector=None, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Bode
        plot (but without the straight line approximations), assumong
        `dt=1`.  fvector specifies the frequencies.  If it is a tuple
        (f1, f2), it sets the frequency limits.  Since a logarithmic
        frequency scale is used, f1 must be greater than 0.

        This method makes the assumption that the expression is causal.

        """
        from .discretetime import dt

        return self.DTFT(causal=True).subs(dt, 1).bode_plot(fvector, **kwargs)

    def nyquist_plot(self, fvector=None, log_frequency=False, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Nyquist
        plot assuming `dt=1`.  fvector specifies the frequencies.  If it is a tuple
        (f1, f2), it sets the frequency limits.

        The unit circle is shown by default.  This can be disabled with `unitcircle=False`.

        `npoints` set the number of plotted points.

        This method makes the assumption that the expression is causal.
        """
        from .discretetime import dt

        if fvector is None:
            fvector = (-0.5, 0.5)
        return self.DTFT(causal=True).subs(dt, 1).nyquist_plot(fvector,
                                                               log_frequency=log_frequency,
                                                               **kwargs)

    def nichols_plot(self, fvector=None, log_frequency=False, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Nichols
        plot assuming `dt=1`.  fvector specifies the frequencies.  If
        it is a tuple (f1, f2), it sets the frequency limits.

        `npoints` set the number of plotted points.

        This method makes the assumption that the expression is causal.

        """
        from .discretetime import dt

        if fvector is None:
            fvector = (-0.5, 0.5)
        return self.DTFT(causal=True).subs(dt, 1).nichols_plot(fvector,
                                                               log_frequency=log_frequency,
                                                               **kwargs)

    def inverse_bilinear_transform(self):

        from .symbols import s
        from .discretetime import dt

        # z = exp(s * dt) gives the exact solution

        return self.subs((1 + s * dt / 2) / (1 - s * dt / 2))

    def DFT(self, N=None, evaluate=True, **assumptions):
        """Determine DFT.

        `N` needs to be a positive integer symbol or a str specifying
        the extent of the DFT.  By default `N` is defined as 'N'."""

        result = self.IZT(**assumptions).DFT(N=N)
        return result

    def discrete_time_fourier_transform(self, var=None, images=oo,
                                        **assumptions):
        """Convert to Fourier domain using discrete time Fourier transform."""
        return self.DTFT(var, images, **assumptions)

    def DTFT(self, var=None, images=oo, **assumptions):
        """Convert to Fourier domain using discrete time Fourier transform."""
        from .symbols import f

        if var is None:
            var = f

        if assumptions.get('causal', self.is_causal):
            result = self.subs(exp(j * 2 * pi * f * dt))
        else:
            result = self.IZT(**assumptions).DTFT(images=images)
        return result(var)

    def fourier(self, var=None, evaluate=True, **assumptions):
        """Attempt discrete-time Fourier transform. This is an alias for DTFT."""

        return self.DTFT(var, evaluate, **assumptions)

    def norm_fourier(self, evaluate=True, **assumptions):
        """Attempt normalized discrete-time Fourier transform."""

        from .symbols import F

        return self.DTFT(F, evaluate, **assumptions)

    def angular_fourier(self, evaluate=True, **assumptions):
        """Attempt angular discrete-time Fourier transform."""

        from .symbols import omega

        return self.DTFT(omega, evaluate, **assumptions)

    def norm_angular_fourier(self, evaluate=True, **assumptions):
        """Attempt normalized angular discrete-time Fourier transform."""

        from .symbols import Omega

        return self.DTFT(Omega, evaluate, **assumptions)

    def as_ab(self):
        """Return lists of denominator and numerator coefficients
        when the denominator and numerator are expressed as polynomials
        in z**-1.  The lowest order coefficients are returned first."""

        from numpy import array

        C, R = self.factor_const()

        zi = symbol('zi')
        H = R.replace(z, 1 / zi).cancel()
        a = H.D.coeffs(zi)
        b = H.N.coeffs(zi)
        return a[::-1], list(array(b) * C)[::-1]

    def as_AB(self):

        C, R = self.factor_const()

        zi = symbol('zi')
        H = R.replace(z, 1 / zi).factor()
        r = Ratfun(H, zi.expr)
        B = ZDomainExpression(r.N).replace(zi, 1 / z)
        A = ZDomainExpression(r.D).replace(zi, 1 / z)

        C1, R1 = A.term_const()
        if C1.is_negative:
            A = -A
            B = -B

        return A, B * C

    def difference_equation(self, inputsym='x', outputsym='y', form='iir'):
        """Create difference equation from transfer function.

        `form` can be 'fir' or 'iir' ('direct form I').
        """

        H = self
        x = nexpr('%s(n)' % inputsym)
        y = nexpr('%s(n)' % outputsym)

        X = x.ZT()
        Y = y.ZT()

        if form in ('iir', 'direct form I'):
            # Direct form I
            return self.dlti_filter().difference_equation()

        elif form == 'fir':
            H = H.partfrac()
            lhs = y
            rhs = (H * X).IZT(causal=True)

        else:
            raise ValueError('Unhandled form ' + form)

        return DifferenceEquation(lhs, rhs, inputsym, outputsym)

    def dlti_filter(self, normalize_a0=True):
        """Create discrete-time linear time-invariant filter from discrete-time
        transfer function."""

        # TODO, perhaps add only to DiscreteTimeDomainTransfer?

        from .dltifilter import DLTIFilter

        if not self.is_rational_function:
            raise ValueError("Not a rational function")

        N = self.N
        D = self.D
        nn = N.coeffs()
        dn = D.coeffs()

        if len(nn) > len(dn):
            raise ValueError("System not causal")

        bn = ExprList((len(dn) - len(nn)) * [0] + nn)
        an = dn

        if normalize_a0:
            bn = [bx / an[0] for bx in bn]
            an = [ax / an[0] for ax in an]

        lpf = DLTIFilter(bn, an)
        return lpf

    def inverse_bilinear_transform(self):
        """Approximate z = exp(s * dt)

        by z = (1 + 0.5 * dt * s) / (1 - 0.5 * dt * s)"""

        # TODO: add frequency warping as an option

        from .symbols import s
        from .discretetime import dt

        a = s * dt / 2

        return self.subs((1 + a) / (1 - a))

    def zdomain(self, **assumptions):
        return self

    def discrete_frequency(self, **assumptions):
        N = assumptions.pop('N', None)
        evaluate = assumptions.pop('evaluate', True)
        return self.IZT(**assumptions).DFT(N, evaluate)

    def discrete_time(self, **assumptions):
        return self.IZT(**assumptions)

    def fourier(self, **assumptions):
        return self.DTFT(**assumptions)

    def angular_fourier(self, **assumptions):
        from .symbols import omega

        return self.DTFT(omega, **assumptions)

    def norm_fourier(self, **assumptions):
        from .symbols import F

        return self.DTFT(F, **assumptions)

    def norm_angular_fourier(self, **assumptions):
        from .symbols_time import Omega

        return self.DTFT(Omega, **assumptions)


def zexpr(arg, **assumptions):
    """Create ZDomainExpression object.  If `arg` is zsym return z"""

    from .expr import Expr

    if arg is zsym:
        return z

    if isinstance(arg, Expr):
        if assumptions == {}:
            return arg
        return arg.__class__(arg, **assumptions)

    return ZDomainExpression(arg, **assumptions)


from .expressionclasses import expressionclasses  # nopep8

classes = expressionclasses.register('Z', ZDomainExpression)
ZDomainVoltage = classes['voltage']
ZDomainCurrent = classes['current']
ZDomainAdmittance = classes['admittance']
ZDomainImpedance = classes['impedance']
ZDomainTransferFunction = classes['transfer']

from .nexpr import DiscreteTimeDomainExpression, nexpr  # nopep8
z = ZDomainExpression('z')



from lcapy import *
from lcapy.phasor import PhasorFrequencyDomainExpression
import unittest


class LcapyTester(unittest.TestCase):
    """Unit tests for lcapy"""

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            pprint(ans1)
            pprint(ans2)
            raise AssertionError(e)

    def test_phasor(self):

        a = phasor(3, omega=7)
        self.assertEqual(a.omega, 7, 'omega')
        self.assertEqual(a.phase, 0, 'phase')
        self.assertEqual(a.magnitude, 3, 'magnitude')
        self.assertEqual(a.N, 3, 'N')
        self.assertEqual(a.D, 1, 'D')

        a = phasor(-3, omega=7)
        self.assertEqual(a.phase, pi, 'phase')
        self.assertEqual(a.magnitude, 3, 'magnitude')

        a = phasor(-3 + 4j, omega=7)
        self.assertEqual(a.magnitude, 5, 'magnitude')

        v = voltage('cos(5 * t)')
        i = current('cos(5 * t)')
        V = v.as_phasor()
        I = i.as_phasor()

        Z = V / I
        self.assertEqual(Z.is_impedance, True, 'impedance')
        self.assertEqual(Z.omega, 5, 'omega')

        V2 = phasorvoltage('cos(5 * t)')
        I2 = phasorcurrent('cos(5 * t)')
        self.assertEqual(V, V2, 'phasorvoltage')
        self.assertEqual(I, I2, 'phasorcurrent')

    def test_phasor_rms(self):

        p = phasor(2)

        self.assertEqual(p.rms(), sqrt(2), 'rms')

    def test_phasor_mul_omega(self):

        p = phasor(2)
        q = p * omega
        r = phasor(2 * omega)

        self.assertEqual(q, r, 'phasor(2) * omega')

    def test_phasor_transform(self):

        H = 1 / s
        P = H(jw)

        self.assertEqual(type(P), PhasorFrequencyDomainExpression, 'H(jw)')

        P = H(j * 2 * pi * f)
        self.assertEqual(
            type(P), PhasorFrequencyDomainExpression, 'H(j * 2 * pi * f)')

        self.assertEqual(P.var, None, 'P.var')



"""This module provides support for Fourier transforms.  It calculates
the bilateral Fourier transform using:

   S(f) = \int_{-\infty}^{\infty} s(t) e^{-j * 2 * \pi * t} dt

It also allows functions that strictly do not have a Fourier transform
by using Dirac deltas.  For example, a, cos(a * t), sin(a * t), exp(j
* a * t).


Copyright 2016--2022 Michael Hayes, UCECE

"""

# TODO:
# Add DiracDelta(t, n)
# Simplify  (-j * DiracDelta(f - 1) + j * DiracDelta(f + 1)).inverse_fourier()
# This should give 2 * sin(2 * pi * t)

from sympy.core.function import AppliedUndef
from sympy import sympify, pi, exp, I, oo, S, sign, sin, cos, sinh, cosh, tanh
from sympy import DiracDelta, Heaviside, FourierTransform, Integral
from sympy import fourier_transform as sympy_fourier_transform, Function
from .sym import symsimplify, j
from .transformer import BilateralForwardTransformer
from .utils import factor_const, similarity_shift, expand_functions
from .extrafunctions import rect, sincn, sincu, trap, tri

__all__ = ('FT', 'IFT')


class FourierTransformer(BilateralForwardTransformer):

    name = 'Fourier transform'

    def key(self, expr, t, f, **kwargs):
        return expr, t, f

    def simplify_term(self, expr, var):

        return symsimplify(expr)

    def sympy(self, expr, t, f):

        result = sympy_fourier_transform(expr, t, f)
        if expr != 0 and result == 0:
            # There is a bug in SymPy where it returns 0.
            self.error()

        if isinstance(result, FourierTransform):
            self.error(
                'Try using partfrac() for partial fraction expansion first.')

        return result

    def func(self, expr, t, f):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        # Convert v(t) to V(f), etc.
        name = expr.func.__name__
        if self.is_inverse:
            func = Function(name[0].lower() + name[1:])
        else:
            func = Function(name[0].upper() + name[1:])

        result = func(f)
        return result

    def integral(self, expr, t, f):

        const, expr = factor_const(expr, t)

        if len(expr.args) != 2:
            self.error()

        integrand = expr.args[0]

        if not isinstance(expr, Integral):
            self.error()

        if len(expr.args[1]) != 3:
            self.error('Require definite integral')

        var = expr.args[1][0]
        limits = expr.args[1][1:]
        const2, expr2 = factor_const(integrand, var)

        if (expr2.is_Function and
                expr2.args[0] == t - var and limits[0] == 0 and limits[1] == oo):
            return const2 * self.term(expr2.subs(t - var, t), t, f) / f

        # Look for convolution integral
        # TODO, handle convolution with causal functions.
        if (limits[0] != -oo) or (limits[1] != oo):
            self.error('Need indefinite limits')

        if ((len(expr.args) != 2) or not expr2.is_Mul or
                not expr2.args[0].is_Function or not expr2.args[1].is_Function):
            self.error('Need integral of product of two functions')

        f1 = expr2.args[0]
        f2 = expr2.args[1]
        # TODO: apply similarity theorem if have f(a * tau) etc.

        if (f1.args[0] == var and f2.args[0] == t - var):
            F1 = self.term(f1, var, f)
            F2 = self.term(f2.subs(t - var, t), t, f)
        elif (f2.args[0] == var and f1.args[0] == t - var):
            F1 = self.term(f1.subs(t - var, t), t, f)
            F2 = self.term(f2, var, f)
        else:
            self.error('Cannot recognise convolution')

        return const2 * F1 * F2

    def function(self, expr, t, f):

        # Handle expressions with a function of FOO, e.g.,
        # v(t), v(t) * y(t),  3 * v(t) / t, v(4 * a * t), etc.,

        if not expr.has(AppliedUndef):
            self.error()

        const, expr = factor_const(expr, t)

        if isinstance(expr, AppliedUndef) and expr.args[0] == t:
            return self.func(expr, t, f) * const

        tsym = sympify(str(t))
        expr = expr.subs(tsym, t)

        rest = S.One
        undefs = []
        for factor in expr.as_ordered_factors():
            if isinstance(factor, AppliedUndef):
                if factor.args[0] != t:
                    self.error('Weird function %s not of %s' % (factor, t))
                undefs.append(factor)
            else:
                rest *= factor

        if rest.has(AppliedUndef):
            # Have something like 1/v(t)
            self.error()

        exprs = undefs
        if rest.has(t):
            exprs = exprs + [rest]
            rest = S.One

        result = self.term(exprs[0], t, f) * rest

        if len(exprs) == 1:
            return result * const

        for m in range(len(exprs) - 1):
            nu = self.dummy_var(expr, 'tau' if self.is_inverse else 'nu',
                                level=m, real=True)
            expr2 = self.term(exprs[m + 1], t, f)
            result = Integral(result.subs(f, f - nu) * expr2.subs(f, nu),
                              (nu, -oo, oo))

        return result * const

    def term(self, expr, t, f):

        const, expr = factor_const(expr, t)

        if expr.has(Integral):
            return self.integral(expr, t, f) * const

        if isinstance(expr, AppliedUndef) and expr.args[0] == t:
            return self.func(expr, t, f) * const

        # TODO add u(t) <-->  delta(f) / 2 - j / (2 * pi * f)

        if expr.has(AppliedUndef) and expr.args[0] == t:
            # Handle v(t), v(t) * y(t),  3 * v(t) / t etc.
            return self.function(expr, t, f) * const

        # Check for constant.
        if not expr.has(t):
            return expr * DiracDelta(f) * const

        one = S.One
        const1 = const
        other = one
        exps = one
        factors = expr.expand().as_ordered_factors()
        for factor in factors:
            if not factor.has(t):
                const1 *= factor
            else:
                if factor.is_Function and factor.func == exp and factor.args[0].has(I):
                    exps *= factor
                else:
                    other *= factor

        sf = -f if self.is_inverse else f

        if other != 1 and exps == 1:
            if other == t:
                return const1 * I / (2 * pi) * DiracDelta(sf, 1)
            elif other == t**2:
                return -const1 / (2 * pi)**2 * DiracDelta(sf, 2)
            # TODO check for other powers of t...
            elif other == sign(t):
                return const1 / (I * pi * sf)
            elif other == sign(t) * t:
                return -const1 * 2 / (2 * pi * f)**2
            elif other == Heaviside(t):
                return const1 / (I * 2 * pi * f) + const1 * DiracDelta(sf) / 2
            elif other == 1 / t:
                return -const1 * I * pi * sign(sf)
            elif other == 1 / t**2:
                return -const1 * 2 * pi**2 * sf * sign(sf)
            elif False and other == exp(t):
                # SymPy does not like Dirac Delta of a complex argument.
                return DiracDelta(sf + I / (2 * pi))
            elif other.is_Function and other.func == Heaviside and other.args[0] == t:
                return (const1 / (I * 2 * pi * sf) + const1 * DiracDelta(sf) / 2)
            elif other == Heaviside(t) * t:
                return -const1 / (2 * pi * f)**2 + const1 * I * DiracDelta(sf, 1) / (4 * pi)
            # t * u(t - tau)
            elif (other.is_Mul and len(other.args) == 2 and
                  other.args[0] == t and other.args[1].is_Function and
                  other.args[1].func == Heaviside and other.args[1].args[0] == t):
                e = exp(I * 2 * pi * sf)
                return I * DiracDelta(sf, 1) / (4 * pi) * e - 1 / (4 * pi**2 * f**2) * e
            # exp(c1 * t + c0) * u(t)
            elif (other.is_Mul and len(other.args) == 2 and
                  other.args[1].is_Function and other.args[1].func == exp and
                  other.args[1].args[0].is_polynomial(t) and
                  other.args[1].args[0].as_poly(t).is_linear and
                  other.args[0].is_Function and
                  other.args[0].func == Heaviside and
                  other.args[0].args[0] == t):
                foo = other.args[1].args[0]
                c0 = foo.coeff(t, 0)
                c1 = foo.coeff(t, 1)
                return const1 * exp(c0) / (I * 2 * pi * sf - c1)
            elif other.is_Function and other.func == sincn and other.args[0] == t:
                return const1 * rect(f)
            elif other.is_Function and other.func == sincu and other.args[0] == t:
                return const1 * pi * rect(f * pi)
            elif (other.is_Pow and other.args[1] == 2 and
                  other.args[0].is_Function and other.args[0].func == sincn and
                  other.args[0].args[0] == t):
                other = other.args[0]
                return const1 * tri(f)
            elif other.is_Function and other.func == rect and other.args[0] == t:
                return const1 * sincn(f)
            elif other.is_Function and other.func == tri and other.args[0] == t:
                return const1 * sincn(f)**2
            elif other.is_Function and other.func == trap and other.args[0] == t:
                alpha = other.args[1]

                # Check for rect
                if alpha == 0:
                    return const1 * sincn(f)
                return alpha * const1 * sincn(f) * sincn(alpha * f)

            #
            # factor = other.factor()
            # const2, factor2 = factor_const(factor, t)
            # if factor2.is_Pow and factor2.args[1] == -2:
            #     foo = factor2.args[0]
            #     a = foo.coeff(t, 1)
            #     b = foo.coeff(t, 0)
            #     if a != 0:
            #         return const1 * const2 * f * exp(-b * 2 *pi * f / a) * sign(f) / (2 * pi)

            # Sympy incorrectly gives exp(-a * t) instead of exp(-a * t) * Heaviside(t)
            elif other.is_Pow and other.args[1] == -1 and other.args[0].has(t):
                foo = other.args[0]

                if (foo.is_polynomial(t) and foo.as_poly(t).is_linear and
                        foo.is_complex):
                    c0 = foo.coeff(t, 0)
                    c1 = foo.coeff(t, 1)
                    s = (2 * pi * I) / c1
                    return const1 * s * exp(c0 * sf * s) * Heaviside(-sf)
                elif foo.is_Function and foo.func == cosh and foo.args[0] == t:
                    return const * pi / cosh(pi**2 * sf)
                elif foo.is_Function and foo.func == sinh and foo.args[0] == t:
                    return -I * const * pi * tanh(pi**2 * sf)
            elif other.is_Function and other.func == tanh and other.args[0] == t:
                return -I * const * pi / sinh(pi**2 * sf)

            if expr == t * DiracDelta(t, 1):
                return const * sf / (-I * 2 * pi)

            # Apply similarity and shift theorems.
            expr2, scale, shift = similarity_shift(expr, t)
            if scale != 1 or shift != 0:
                result = self.term(expr2, t, f / scale) / abs(scale)
                if shift != 0:
                    result *= exp(-I * 2 * pi * f / scale * shift)
                return const * result

            if expr.is_Function and expr.args[0] == t:
                # Try to handle functions such as ramp, rampstep.
                expr2 = expand_functions(expr, t)
                if expr != expr2:
                    terms = expr2.as_ordered_terms()
                    result = 0
                    for term in terms:
                        result += self.term(term, t, f)
                    return result * const

            # Punt and use SymPy.  Should check for t**n, t**n * exp(-a * t), etc.
            return const * self.sympy(expr, t, sf)

        args = exps.args[0]
        foo = args / t
        if foo.has(t):
            # Have exp(a * t**n), SymPy might be able to handle this
            return const * self.sympy(expr, t, sf)

        if exps != 1 and foo.has(I):
            if other == 1:
                return const1 * DiracDelta(sf - foo / (I * 2 * pi))
            Q = self.term(other, t, f)
            return const1 * Q.subs(f, (f - foo / (I * 2 * pi)))

        return const * self.sympy(expr, t, sf)

    def noevaluate(self, expr, t, f):

        return Integral(expr * exp(-j * 2 * pi * f * t), (t, -oo, oo))

    def check(self, expr, t, f):

        if expr.has(f):
            self.error('Expression depends on f')

        if expr.is_Piecewise and expr.args[0].args[1].has(t >= 0):
            self.error('Expression is unknown for t < 0 (use causal=True)')

    def rewrite(self, expr, var):

        # sym.rewrite(exp) can create exp(log...)
        if expr.has(sin):
            expr = expr.replace(lambda expr: expr.is_Function and expr.func == sin,
                                lambda expr: expr.rewrite(exp))
        if expr.has(cos):
            expr = expr.replace(lambda expr: expr.is_Function and expr.func == cos,
                                lambda expr: expr.rewrite(exp))

        expr = expr.expand()
        return expr


fourier_transformer = FourierTransformer()


def FT(expr, t, f, **kwargs):
    """Compute bilateral Fourier transform of expr.

    Undefined functions such as v(t) are converted to V(f)

    This also handles some expressions that do not really have a Fourier
    transform, such as a, cos(a * t), sin(a * t), exp(I * a * t)."""

    return fourier_transformer.transform(expr, t, f, **kwargs)


def fourier_transform(expr, t, f, **kwargs):
    """Compute bilateral Fourier transform of expr.

    Undefined functions such as v(t) are converted to V(f)

    This also handles some expressions that do not really have a Fourier
    transform, such as a, cos(a * t), sin(a * t), exp(I * a * t)."""

    return fourier_transformer.transform(expr, t, f, **kwargs)



from lcapy import *
from lcapy.sexpr import LaplaceDomainVoltage
from lcapy.texpr import TimeDomainExpression, TimeDomainVoltage
import unittest
import sympy as sym


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_laplace(self):

        self.assertEqual(Heaviside(t).laplace(), 1 / s, "Heaviside(t)")
        self.assertEqual(DiracDelta(t).laplace(), 1, "DiracDelta(t)")
        self.assertEqual(TimeDomainVoltage('x(t)').laplace(),
                         LaplaceDomainVoltage('X(s)'), "x(t)")
        self.assertEqual(TimeDomainVoltage('5 * x(t)').laplace(),
                         LaplaceDomainVoltage('5 * X(s)'), "5 * x(t)")

        v = expr('R0 * exp(-alpha * t) * i(t)')
        V = expr('R0 * I(s + alpha)')

        self.assertEqual(v.laplace(), V, "R0 * exp(-alpha * t) * i(t)")

        self.assertEqual(rect(t - 0.5).laplace(), (1 - exp(-s)) / s, "rect(t)")
        self.assertEqual(rect(t / 2 - 0.5).laplace(),
                         (1 - exp(-2 * s)) / s, "rect(t / 2)")
        self.assertEqual(ramp(t).laplace(), 1 / s**2, "ramp(t)")
        self.assertEqual(rampstep(t).laplace(),
                         (1 - exp(-s)) / s**2, "rampstep(t)")

    def test_inverse_laplace(self):

        self.assertEqual((1 / s).inverse_laplace(causal=True), Heaviside(t),
                         "1 / s")
        self.assertEqual((s * 0 + 1).inverse_laplace(causal=True), DiracDelta(t),
                         "1")
        self.assertEqual((s).inverse_laplace(causal=True), DiracDelta(t, 1),
                         "s")
        self.assertEqual((s * 0 + 10).inverse_laplace(causal=True), 10
                         * DiracDelta(t), "10")
        self.assertEqual(LaplaceDomainVoltage('V(s)').inverse_laplace(causal=True),
                         TimeDomainVoltage('v(t)'), "V(s)")
        self.assertEqual(LaplaceDomainVoltage('10 * V(s)').inverse_laplace(causal=True),
                         TimeDomainVoltage('10 * v(t)'), "V(s)")
        self.assertEqual(LaplaceDomainVoltage('10 * V(s) * exp(-5 * s)').inverse_laplace(
            causal=True), TimeDomainVoltage('10 * v(t - 5)'), "10 * V(s) * exp(-5 * s)")
        self.assertEqual(TimeDomainVoltage('v(t)').laplace().inverse_laplace(causal=True),
                         TimeDomainVoltage('v(t)'), "v(t)")
        self.assertEqual(expr('1/(s+a)').inverse_laplace(causal=True),
                         expr('exp(-a * t) * u(t)'), "1/(s+a)")
        self.assertEqual(
            expr('1/(s**2)').inverse_laplace(causal=True), expr('t * u(t)'), "1/(s**2)")
        self.assertEqual(expr('1/((s+3)**2)').inverse_laplace(causal=True),
                         expr('t * u(t) * exp(-3 * t)'), "1/((s+3)**2)")
        self.assertEqual(expr('1/(s**3)').inverse_laplace(causal=True),
                         expr('t**2 * u(t) / 2'), "1/(s**3)")
        self.assertEqual(expr('s/(s+a)').inverse_laplace(causal=True),
                         expr('-a * exp(-a * t) * u(t) + delta(t)'), "s/(s+a)")
        self.assertEqual(expr('s/(s**2+a**2)').inverse_laplace(causal=True),
                         expr('cos(a * t) * u(t)'), "s/(s**2+a**2)")
        self.assertEqual(expr('a/(s**2+a**2)').inverse_laplace(causal=True),
                         expr('sin(a * t) * u(t)'), "a/(s**2+a**2)")
        self.assertEqual(expr('(1/s - exp(-s)/s)/s').ILT(causal=True),
                         expr('(t*u(t) + (1 - t)*u(t - 1))'), "(1/s - exp(-s)/s)/s")
        self.assertEqual(expr('(1/s - exp(-s)/s)/s').expand().ILT(causal=True),
                         expr('(t*u(t) + (1 - t)*u(t - 1))'), "(1/s - exp(-s)/s)/s")

        self.assertEqual(expr('s*(1/s - exp(-5*s)/s)/(s + 1)').ILT(causal=True),
                         expr('-exp(5 - t)*u(t - 5) + exp(-t)*u(t)'),
                         "'s*(1/s - exp(-5*s)/s)/(s + 1)")

        self.assertEqual(expr('s*(1/s - exp(-5*s)/s)/(s + 1)').expand().ILT(causal=True),
                         expr('-exp(5 - t)*u(t - 5) + exp(-t)*u(t)'),
                         "'s*(1/s - exp(-5*s)/s)/(s + 1)")

        A4 = (1/s - exp(-5*s)/s)/(s**2 + 707*s/500 + 1)
        self.assertEqual(A4(t)(s), A4, str(A4))
        self.assertEqual(expr('a/(s+a)', causal=True)(t),
                         expr('a*exp(-a * t)*u(t)'), "s/(s+a)")

    def test_damped_sin(self):

        H1 = 2 / (2 * s ** 2 + 5 * s + 6)
        H2 = H1 * s
        H3 = H1 * s * s

        self.assertEqual(H1(t, damped_sin=True)(s), H1, "damped sin1")
        self.assertEqual(H2(t, damped_sin=True)(s), H2, "damped sin2")
        self.assertEqual(H3(t, damped_sin=True)(s), H3, "damped sin3")
        self.assertEqual((H1 + H2)(t, damped_sin=True)
                         (s), H1 + H2, "damped sin1, 2")
        self.assertEqual((H1 + H3)(t, damped_sin=True)
                         (s), H1 + H3, "damped sin1, 3")
        self.assertEqual((H1 + H2 + H3)(t, damped_sin=True)
                         (s), H1 + H2 + H3, "damped sin1, 2, 3")

    def test_derivative_undef(self):

        H = s * 'I(s)' - 'i(0)'
        h = H(t)
        H2 = h(s)

        self.assertEqual(H, H2, "derivative of undef")

        H = s**2 * 'I(s)'
        h = H(t)
        H2 = h(s)

        self.assertEqual(H, H2, "second derivative of undef")

    def test_laplace_convolution(self):

        a = expr('Integral(3 * x(t - tau) * y(tau), (tau, -oo, oo))')
        self.assertEqual(a(s), expr('3 * X(s) * Y(s)'), "3 * X * Y")
        a = expr('Integral(3 * x(tau) * y(t - tau), (tau, -oo, oo))')
        self.assertEqual(a(s), expr('3 * X(s) * Y(s)'), "3 * X * Y")
        a = expr('Integral(3 * x(t - tau) * Heaviside(tau), (tau, -oo, oo))')
        self.assertEqual(a(s), expr('3 * X(s) / s'), "3 * X / s")
        a = expr('Integral(3 * x(tau) * Heaviside(t - tau), (tau, -oo, oo))')
        self.assertEqual(a(s), expr('3 * X(s) / s'), "3 * X / s")

        a = expr('Integral(3 * x(t - tau), (tau, 0, oo))')
        self.assertEqual(a(s), expr('3 * X(s) / s'), "3 * X / s")

        a = expr('Integral(3 * x(tau), (tau, 0, t))')
        self.assertEqual(a(s), expr('3 * X(s) / s'), "3 * X / s")

        a = expr('x(t)').convolve(expr('g(t)'))
        self.assertEqual(a(s), expr('X(s) * G(s)'), "X(s) * G(s)")

        a = expr('x(t)').convolve(expr('g(t)'), commutate=True)
        self.assertEqual(a(s), expr('X(s) * G(s)'), "X(s) * G(s)")

        a = expr('x(t)').convolve(expr('g(t)', causal=True))
        self.assertEqual(a(s), expr('X(s) * G(s)'), "X(s) * G(s)")

        a = expr('x(t)').convolve(expr('g(t)', causal=True), commutate=True)
        self.assertEqual(a(s), expr('X(s) * G(s)'), "X(s) * G(s)")

        a = expr('x(t)', causal=True).convolve(expr('g(t)'))
        self.assertEqual(a(s), expr('X(s) * G(s)'), "X(s) * G(s)")

        a = expr('x(t)', causal=True).convolve(expr('g(t)', commutate=True))
        self.assertEqual(a(s), expr('X(s) * G(s)'), "X(s) * G(s)")

        a = expr('x(t)', causal=True).convolve(expr('g(t)', causal=True))
        self.assertEqual(a(s), expr('X(s) * G(s)'), "X(s) * G(s)")

        a = expr('x(t)', causal=True).convolve(expr('g(t)', causal=True,
                                                    commutate=True))
        self.assertEqual(a(s), expr('X(s) * G(s)'), "X(s) * G(s)")



"""This module provides the DiscreteFourierDomainExpression class to
 represent k-domain (discrete Fourier domain) expressions.

Copyright 2020--2021 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import DiscreteFourierDomain
from .inverse_fourier import inverse_fourier_transform
from .functions import exp
from .sym import j, oo, pi
from .sym import nsym, ksym, zsym
from .seqexpr import SequenceExpression
from .kseq import DiscreteFourierDomainSequence
from .inverse_dft import IDFT
from sympy import Sum

__all__ = ('kexpr', )


class DiscreteFourierDomainExpression(DiscreteFourierDomain, SequenceExpression):
    """Discrete Fourier domain expression or symbol."""

    var = ksym
    seqcls = DiscreteFourierDomainSequence

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)
        if 'integer' not in assumptions:
            assumptions['real'] = True

        super(DiscreteFourierDomainExpression,
              self).__init__(val, **assumptions)

        expr = self.expr
        if check and expr.has(zsym):
            raise ValueError(
                'k-domain expression %s cannot depend on z' % expr)
        if check and expr.has(nsym) and not expr.has(Sum):
            raise ValueError(
                'k-domain expression %s cannot depend on n' % expr)

    def as_expr(self):
        return DiscreteFourierDomainExpression(self)

    def plot(self, kvector=None, **kwargs):
        """Plot frequency response at values specified by kvector.  If kvector
        is a tuple, this sets the frequency limits.

        kwargs include:
        axes - the plot axes to use otherwise a new figure is created
        xlabel - the x-axis label
        ylabel - the y-axis label
        ylabel2 - the second y-axis label if needed, say for mag and phase
        xscale - the x-axis scaling, say for plotting as ms
        yscale - the y-axis scaling, say for plotting mV
        plot_type -  'dB_phase', 'mag-phase', 'real-imag', 'mag', 'phase',
        'real', or 'imag'
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned.

        There are many plotting options, see lcapy.plot and
        matplotlib.pyplot.plot.

        For example:
            V.plot(kvector, log_frequency=True)
            V.real.plot(kvector, color='black')
            V.phase.plot(kvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of magnitude (dB)
        and phase.

        """

        if kvector is None:
            kvector = (0, 20)

        from .plot import plot_sequence
        return plot_sequence(self, kvector, **kwargs)

    def IDFT(self, N=None, evaluate=True):
        """Determine inverse DFT.

        `N` needs to be a positive integer symbol or a str specifying
        the extent of the inverse DFT.  By default `N` is defined as
        'N'."""

        from .sym import miscsymbol

        if N is None:
            N = miscsymbol('N', integer=True, positive=True)
        elif isinstance(N, str):
            N = miscsymbol(N, integer=True, positive=True)

        result = IDFT(self.expr, ksym, nsym, N, evaluate=evaluate)
        result = self.change(result, domain='discrete time')
        result = result.simplify_unit_impulse()
        return result

    def ZT(self, **assumptions):
        N = assumptions.pop('N', None)
        evaluate = assumptions.pop('evaluate', True)
        return self.IDFT(N=N).ZT(evaluate=evaluate)

    def zdomain(self, **assumptions):
        return self.ZT(**assumptions)

    def discrete_frequency(self, **assumptions):
        return self

    def discrete_time(self, **assumptions):
        return self.IDFT(**assumptions)

    def fourier(self, **assumptions):
        return self.IDFT(**assumptions).DTFT(**assumptions)

    def angular_fourier(self, **assumptions):
        from .symbols import omega

        return self.IDFT(**assumptions).DTFT(omega, **assumptions)

    def norm_fourier(self, **assumptions):
        from .symbols import F

        return self.IDFT(**assumptions).DTFT(F, **assumptions)

    def norm_angular_fourier(self, **assumptions):
        from .symbols_time import Omega

        return self.IDFT(**assumptions).DTFT(Omega, **assumptions)


def kexpr(arg, **assumptions):
    """Create kExpr object.  If `arg` is ksym return k"""

    from .expr import Expr

    if arg is ksym:
        return k

    if isinstance(arg, Expr):
        if assumptions == {}:
            return arg
        return arg.__class__(arg, **assumptions)

    return DiscreteFourierDomainExpression(arg, **assumptions)


from .expressionclasses import expressionclasses  # nopep8

expressionclasses.register('discrete fourier', DiscreteFourierDomainExpression)

k = DiscreteFourierDomainExpression('k', integer=True)



from .quantity import Quantity


class CurrentSquaredMixin(Quantity):

    quantity = 'currentsquared'
    quantity_label = 'Current^2'
    quantity_units = 'A^2'
    is_currentsquared = True
    is_squared = True



from lcapy import *
import unittest


class LcapyCoreTester(unittest.TestCase):
    """Unit tests for lcapy core"""

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            pprint(ans1)
            pprint(ans2)
            raise AssertionError(e)

    def test_simplify_sin_cos(self):
        """Test simplify_sin_cos"""

        a = expr('3 * cos(theta) + 4 * sin(theta)')
        b = expr('5 * cos(theta - atan(4 / 3))')

        # Note, simplify converts the results to the same form.
        self.assertEqual(a.simplify_sin_cos(
            as_cos=True).simplify(), b, "cos form")
        self.assertEqual(a.simplify_sin_cos(
            as_sin=True).simplify(), b, "sin form")
        self.assertEqual(a.simplify_sin_cos().simplify(), b, "default form")

        a = 3 * sin(2 * t) + 4 * cos(2 * t) + 5 * sin(4 * t) + \
            12 * cos(4 * t) + cos(2 * t + 1) + 3
        b = 5 * cos(2 * t - atan('3 / 4')) + 13 * \
            cos(4 * t - atan('5 / 12')) + cos(2 * t + 1) + 3

        self.assertEqual(a.simplify_sin_cos().simplify(), b, "default")

    def test_simplify_heaviside(self):
        """Test simplify_heaviside"""

        a = Heaviside(4 * t - 2)
        b = a.simplify_heaviside()
        self.assertEqual(b, Heaviside(t - 0.5), "simplify_heaviside scale")

        a = Heaviside(t) * Heaviside(t)
        self.assertEqual(a.simplify_heaviside(), Heaviside(t),
                         "simplify_heaviside product")

    def test_simplify_rect(self):
        """Test simplify_rect"""

        a = rect(t) * rect(t)
        self.assertEqual(a.simplify_rect(), rect(t), "simplify_rect product")

    def test_simplify_dirac_delta(self):
        """Test simplify_dirac_delta"""

        a = DiracDelta(t) * 'g(t)'
        self.assertEqual(a.simplify_dirac_delta(), DiracDelta(
            t) * 'g(0)', "simplify_dirac_delta product")

        a = DiracDelta(3 * t)
        self.assertEqual(a.simplify_dirac_delta(), DiracDelta(
            t) / 3, "simplify_dirac_delta scale")

    def test_simplify_conjugates(self):
        """Test simplify_conjugates"""

        a = exp(j * 2) + exp(-j * 2) + exp(j * 3)
        self.assertEqual(a.simplify_conjugates(), 2 * cos(2) + exp(j * 3), "")

    def test_expand_hyperbolic_trig(self):
        """Test expand_hyperbolic_trig"""

        a = cosh(2) + sinh(2)
        self.assertEqual(a.expand_hyperbolic_trig(), 2 * exp(2), "")

        a = cosh(2) - sinh(2)
        self.assertEqual(a.expand_hyperbolic_trig(), 2 * exp(-2), "")



"""This module defines the domain classes.

Copyright 2021 Michael Hayes, UCECE

"""

from .units import u as uu


class Domain(object):
    is_undefined_domain = False
    is_constant_domain = False
    is_time_domain = False
    is_laplace_domain = False
    is_fourier_domain = False
    is_angular_fourier_domain = False
    is_phasor_domain = False
    is_phasor_time_domain = False
    is_phasor_frequency_domain = False
    is_fourier_noise_domain = False
    is_angular_fourier_noise_domain = False
    is_discrete_time_domain = False
    is_discrete_fourier_domain = False
    is_Z_domain = False
    is_transform_domain = False
    is_superposition_domain = False
    is_one_sided = False
    is_sequence = False


class UndefinedDomain(Domain):
    domain = 'undefined'
    domain_label = 'Undefined'
    domain_units = 1
    is_undefined_domain = True


class ConstantDomain(Domain):
    domain = 'constant'
    domain_label = 'Constant'
    domain_units = 1
    is_constant_domain = True


class TimeDomain(Domain):
    domain = 'time'
    domain_label = 'Time'
    domain_units = uu.s
    is_time_domain = True


class LaplaceDomain(Domain):
    domain = 'laplace'
    domain_label = 'Laplace frequency'
    domain_units = uu.rad / uu.s
    is_laplace_domain = True
    is_transform_domain = True


class FourierDomain(Domain):
    domain = 'fourier'
    domain_label = 'Frequency'
    domain_units = uu.Hz
    is_fourier_domain = True
    is_transform_domain = True


class NormFourierDomain(Domain):
    domain = 'norm fourier'
    domain_label = 'Normalized frequency'
    domain_units = 1
    is_norm_fourier_domain = True
    is_transform_domain = True


class AngularFourierDomain(Domain):
    domain = 'angular fourier'
    domain_label = 'Angular frequency'
    domain_units = uu.rad / uu.s
    is_angular_fourier_domain = True
    is_transform_domain = True


class NormAngularFourierDomain(Domain):
    domain = 'norm angular fourier'
    domain_label = 'Normalized angular frequency'
    domain_units = uu.rad
    is_norm_angular_fourier_domain = True
    is_transform_domain = True


class PhasorTimeDomain(Domain):
    domain = 'phasor'
    domain_label = ''
    domain_units = 1
    is_phasor_domain = True
    is_transform_domain = True


class PhasorFrequencyDomain(Domain):
    domain = 'phasor'
    domain_label = 'Angular Frequency'
    domain_units = uu.rad / uu.s
    is_phasor_domain = True
    is_transform_domain = True


class FourierNoiseDomain(Domain):
    domain = 'fourier noise'
    domain_label = 'Frequency'
    domain_units = uu.Hz
    is_fourier_noise_domain = True
    is_transform_domain = True
    is_one_sided = True


class AngularFourierNoiseDomain(Domain):
    domain = 'angular fourier noise'
    domain_label = 'Angular frequency'
    domain_units = uu.rad / uu.s
    is_angular_fourier_noise_domain = True
    is_transform_domain = True
    is_one_sided = True


class DiscreteTimeDomain(Domain):
    domain = 'discrete time'
    domain_label = 'Discrete time'
    domain_units = 1
    is_discrete_time_domain = True
    is_transform_domain = False


class DiscreteFourierDomain(Domain):
    domain = 'discrete fourier'
    domain_label = 'Discrete Fourier'
    domain_units = 1
    is_discrete_fourier_domain = True
    is_transform_domain = True


class ZDomain(Domain):
    domain = 'Z'
    domain_label = 'Z'
    domain_units = 1
    is_Z_domain = True
    is_transform_domain = True


class SuperpositionDomain(Domain):
    domain = 'superposition'
    domain_label = 'Superposition'
    domain_units = 1
    is_superposition_domain = True
    is_transform_domain = True



from lcapy import *
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_DFT(self):

        self.assertEqual(nexpr('delta(n)').DFT(), 1, "delta(n)")
        self.assertEqual(nexpr('2 * delta(n)').DFT(), 2, "2 * delta(n)")

        self.assertEqual(nexpr('1').DFT(), kexpr('N * delta(k)'), "1")
        self.assertEqual(nexpr('2').DFT(), kexpr('2 * N * delta(k)'), "2")

        self.assertEqual(nexpr('x(n)').DFT(), kexpr('X(k)'), "x(n)")
        self.assertEqual(nexpr('2 * x(n)').DFT(),
                         kexpr('2 * X(k)'), "2 * x(n)")
        self.assertEqual(nexpr('x(2 * n)').DFT(),
                         kexpr('X(k / 2) / 2'), "x(2 * n)")

        self.assertEqual(nexpr('delta(n - 1)').DFT(),
                         kexpr('exp(-j * 2 * pi * k / N)'), "delta(n - 1)")

        self.assertEqual(nexpr('2 * delta(n - 1)').DFT(),
                         kexpr('2 * exp(-j * 2 * pi * k / N)'), "2 * delta(n - 1)")

        self.assertEqual(nexpr('delta(n - 2)').DFT(),
                         kexpr('exp(-j * 2 * pi * 2 * k / N)'), "delta(n - 2)")

        # Temp. hack, ideally should get result 2 * N * delta(k + 1)
        self.assertEqual(nexpr('2 * exp(-j * 2 * pi * n / N)').DFT(),
                         kexpr('2 * N * delta(k + 1 - N)'), "2 * exp(-j * 2 * pi * n / N)")
        self.assertEqual(nexpr('x(n-1)').DFT(N=2),
                         kexpr('X(k)*exp(-j * pi * k)'), "x(n-1)")

    def test_IDFT(self):

        self.assertEqual(kexpr('X(k)').IDFT(), nexpr('x(n)'), "X(k)")
        self.assertEqual(kexpr(1).IDFT(), delta(n), "1")
        self.assertEqual(delta(k).IDFT(), nexpr('1 / N'), "delta(k)")

    def test_DFT_matrix(self):

        X1 = DFTmatrix(4)
        X2 = Matrix(((1, 1, 1, 1), (1, -j, -1, j),
                    (1, -1, 1, -1), (1, j, -1, -j)))
        X3 = IDFTmatrix(4).conj * 4

        self.assertEqual(X1, X2, "DFTmatrix(4)")
        self.assertEqual(X3, X2, "IDFTmatrix(4)")
